from typing import Optional

from django.utils.translation import gettext as _
from django.utils.translation import pgettext
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbcrm.models import Activity


def get_activity_legend() -> list[dp.Legend]:
    """Dynamically creates the activity legend based on the activity status enum"""

    legend_items = []
    for status, color in Activity.Status.get_color_map():
        legend_items.append(dp.LegendItem(icon=color, label=status.label, value=status.value))
    return [dp.Legend(key="status", items=legend_items)]


def get_activity_list_formatting() -> list[dp.Formatting]:
    """Dynamically creates the activity list formatting based on the activity status enum"""

    formatting_rules = []
    for status, color in Activity.Status.get_color_map():
        formatting_rules.append(dp.FormattingRule(condition=("==", status.name), style={"backgroundColor": color}))

    return [dp.Formatting(column="status", formatting_rules=formatting_rules)]


class ActivityTypeDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="icon", label=_("Icon")),
                dp.Field(key="color", label=_("Color")),
                dp.Field(key="score", label=_("Multiplier")),
                dp.Field(key="default", label=_("Is Default")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["default", "."], ["title", "score"], ["color", "icon"]])


class ActivityDisplay(DisplayViewConfig):
    @classmethod
    def _get_activity_instance_display(cls, instance: Activity) -> Display:
        """Returns an activity instance's display

        Args:
            instance: The activity instance that will be displayed

        Returns:
            Display: The display instance
        """

        participants_section = Section(
            key="participants_section",
            collapsible=False,
            title=_("Participants"),
            display=Display(
                pages=[
                    Page(
                        title=_("Participants"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["participants_table"]],
                                inlines=[Inline(key="participants_table", endpoint="activity_participants_table")],
                            )
                        },
                    ),
                ]
            ),
        )

        if instance.status == Activity.Status.PLANNED:
            display = Display(
                pages=[
                    Page(
                        title=_("Main Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[
                                    [
                                        "status",
                                        "status",
                                        "participants_section",
                                        "participants_section",
                                    ],
                                    [
                                        "title",
                                        "type",
                                        "participants_section",
                                        "participants_section",
                                    ],
                                    [
                                        "period",
                                        "all_day",
                                        "participants_section",
                                        "participants_section",
                                    ],
                                    [
                                        "participants",
                                        "groups",
                                        "participants_section",
                                        "participants_section",
                                    ],
                                    [
                                        "companies",
                                        "disable_participant_check",
                                        "participants_section",
                                        "participants_section",
                                    ],
                                    [
                                        "description",
                                        "description",
                                        "result",
                                        "result",
                                    ],
                                    ["summary", "summary", "summary", "summary"],
                                ],
                                grid_auto_columns="1fr",
                                grid_auto_rows=Style.MIN_CONTENT,
                                sections=(participants_section,),
                            ),
                        },
                    ),
                    Page(
                        title=_("Recurrence"),
                        layouts={
                            default(): Layout(grid_template_areas=[["propagate_for_all_children", "repeat_choice"]])
                        },
                    ),
                    Page(
                        title=_("Additional Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[
                                    [
                                        "assigned_to",
                                        "assigned_to",
                                        "reminder_choice",
                                        "visibility",
                                    ],
                                    ["conference_room", "online_meeting", "location", "importance"],
                                    [
                                        "created",
                                        "edited",
                                        ".",
                                        ".",
                                    ],
                                ],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 2fr)",
                                    "minmax(min-content, 2fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                            )
                        },
                    ),
                ]
            )
        else:
            display = Display(
                pages=[
                    Page(
                        title=_("Main Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[
                                    [
                                        "status",
                                        ".",
                                    ],
                                    [
                                        "title",
                                        "type",
                                    ],
                                    [
                                        "period",
                                        "all_day",
                                    ],
                                    [
                                        "participants",
                                        "groups",
                                    ],
                                    [
                                        "companies",
                                        "disable_participant_check",
                                    ],
                                    [
                                        "description",
                                        "description" if instance.status == Activity.Status.CANCELLED else "result",
                                    ],
                                    ["summary", "summary"],
                                ],
                                grid_auto_columns="1fr",
                                grid_auto_rows=Style.MIN_CONTENT,
                            ),
                        },
                    ),
                    Page(
                        title=_("Participant Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["participants_section"]], sections=(participants_section,)
                            )
                        },
                    ),
                    Page(
                        title=_("Recurrence"),
                        layouts={default(): Layout(grid_template_areas=[["propagate_for_all_children"]])},
                    ),
                    Page(
                        title=_("Additional Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[
                                    [
                                        "assigned_to",
                                        "assigned_to",
                                        "reminder_choice",
                                        "visibility",
                                    ],
                                    ["conference_room", "online_meeting", "location", "importance"],
                                    [
                                        "created",
                                        "edited",
                                        ".",
                                        ".",
                                    ],
                                ],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 2fr)",
                                    "minmax(min-content, 2fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                            )
                        },
                    ),
                ]
            )

        return display

    @classmethod
    def _get_new_activity_instance_display(cls) -> Display:
        """Returns the display for creating a new activity

        Returns:
            Display: The display instance
        """

        new_recurrence_section = Section(
            key="new_recurrence_section",
            collapsible=False,
            title=_("Repeat Until"),
            display=Display(
                pages=[
                    Page(
                        title=_("Repeat Until"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["recurrence_end", "recurrence_count", "."]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 1fr)",
                                    "minmax(min-content, 2fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                            )
                        },
                    ),
                ]
            ),
        )

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                [
                                    "title",
                                    "type",
                                ],
                                [
                                    "period",
                                    "all_day",
                                ],
                                [
                                    "participants",
                                    "groups",
                                ],
                                [
                                    "companies",
                                    "disable_participant_check",
                                ],
                                [
                                    "description",
                                    "result",
                                ],
                            ],
                            grid_auto_columns="1fr",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Recurrence"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["repeat_choice", "."],
                                ["new_recurrence_section", "new_recurrence_section"],
                            ],
                            grid_template_columns=["minmax(min-content, 1fr)", "minmax(min-content, 3fr)"],
                            grid_auto_rows=Style.MIN_CONTENT,
                            sections=(new_recurrence_section,),
                        )
                    },
                ),
                Page(
                    title=_("Additional Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                [
                                    "assigned_to",
                                    "assigned_to",
                                    "reminder_choice",
                                    "visibility",
                                ],
                                ["conference_room", "online_meeting", "location", "importance"],
                            ],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 2fr)",
                                "minmax(min-content, 2fr)",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                        )
                    },
                ),
            ]
        )

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="heat", label=_("Sentiment"), width=75),
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="summary", label=_("Summary")),
                dp.Field(key="period", label=_("Period")),
                dp.Field(key="participants", label=_("Participants")),
                dp.Field(key="companies", label=_("Companies")),
                dp.Field(key="groups", label=_("Groups")),
                dp.Field(key="edited", label=_("Edited")),
                dp.Field(key="created", label=pgettext("As a table header", "Created")),
                dp.Field(key="description", label=_("Description")),
                dp.Field(key="result", label=pgettext("As a table header", "Review")),
                dp.Field(key="latest_reviewer", label=_("Latest Reviewer")),
            ],
            legends=get_activity_legend(),
            formatting=get_activity_list_formatting(),
        )

    def get_instance_display(self) -> Display:
        return (
            self._get_activity_instance_display(self.view.get_object())
            if "pk" in self.view.kwargs
            else self._get_new_activity_instance_display()
        )


class ActivityParticipantDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="participant", label=_("Participant")),
                dp.Field(key="participation_status", label=_("Participation Status")),
                dp.Field(key="status_changed", label=_("Status Changed")),
                dp.Field(key="position", label=_("Position")),
                dp.Field(key="primary_email", label=_("Primary Email")),
                dp.Field(key="primary_telephone", label=_("Primary Phone Number")),
            ],
            legends=[
                dp.Legend(
                    key="is_occupied_filter",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Is Available"),
                            value=False,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=_("Is Occupied"),
                            value=True,
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="is_occupied",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", False),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["participation_status", "participation_status"], ["participant", "customer_status"]]
        )



import functools
from contextlib import suppress
from urllib.parse import urlencode

from django.dispatch import receiver
from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.directory.serializers import (
    CompanyModelSerializer,
    EntryModelSerializer,
    EntryRepresentationSerializer,
    PersonModelListSerializer,
    PersonModelSerializer,
    TelephoneContactSerializer,
)
from wbcore.signals import add_additional_resource, add_instance_additional_resource

from wbcrm.models import ActivityType


@functools.lru_cache()
def get_call_activity_type() -> int:
    return ActivityType.objects.get_or_create(slugify_title="call", defaults={"title": "Call"})[0].id


@receiver(add_additional_resource, sender=TelephoneContactSerializer)
def add_telephone_contact_activity_resources(sender, serializer, instance, request, user, **kwargs):
    res = {}
    with suppress(Entry.DoesNotExist):
        if entry := instance.entry:
            activity_reverse_url = reverse("wbcrm:activity-list", args=[], request=request)

            # Creates the URL for the 'Create New Call Activity'-Button
            query_args = {
                "type": get_call_activity_type(),
                "new_mode": True,
                "participants": [str(request.user.profile.id)],
                "title": _("Call with {name}").format(name=entry.computed_str),
            }

            if entry.is_company:
                query_args["companies"] = [str(entry.id)]
                activity_reverse_url = f"{activity_reverse_url}?companies={entry.id}"
            else:
                query_args["participants"].append(str(entry.id))
                activity_reverse_url = f"{activity_reverse_url}?participants={entry.id}"

            query_args["participants"] = ",".join(query_args["participants"])
            if "companies" in query_args:
                query_args["companies"] = ",".join(query_args["companies"])
        res["list_of_activities"] = activity_reverse_url
        res["new_call"] = reverse("wbcrm:activity-list", args=[], request=request) + "?" + urlencode(query_args)
    return res


@receiver(add_instance_additional_resource, sender=CompanyModelSerializer)
@receiver(add_instance_additional_resource, sender=PersonModelSerializer)
@receiver(add_instance_additional_resource, sender=EntryModelSerializer)
@receiver(add_instance_additional_resource, sender=PersonModelListSerializer)
@receiver(add_instance_additional_resource, sender=EntryRepresentationSerializer)
def add_entry_additional_resources(sender, serializer, instance, request, user, **kwargs):
    res = {"account": f'{reverse("wbcrm:account-list", args=[], request=request)}?customer={instance.id}'}
    if instance.is_company:
        res["activity"] = f'{reverse("wbcrm:activity-list", request=request)}?companies={instance.id}'
        res["interested_products"] = reverse(
            "wbcrm:company-interestedproduct-list", args=[instance.id], request=request
        )
    else:
        res["activity"] = f'{reverse("wbcrm:activity-list", request=request)}?participants={instance.id}'
    return res



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ChildAccountAccountEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcrm:account-childaccount-list",
            args=[self.view.kwargs["account_id"]],
            request=self.request,
        )


class AccountRoleAccountEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcrm:account-accountrole-list",
            args=[self.view.kwargs["account_id"]],
            request=self.request,
        )


class InheritedAccountRoleAccountEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class AccountRoleValidityEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcrm:accountrole-validity-list",
            args=[self.view.kwargs["account_role_id"]],
            request=self.request,
        )



from .accounts import AccountFilter, AccountRoleFilterSet
from .activities import (
    ActivityChartFilter,
    ActivityFilter,
    ActivityParticipantFilter,
    ActivityTypeFilter,
)
from .groups import GroupFilter
from .products import ProductCompanyFilterSet, ProductFilterSet
from .signals import *



import copy

import pytest
from django.contrib.messages import storage
from rest_framework.test import APIRequestFactory


class MSGraphFixture:
    token = None
    _subscriptions = []
    _subscription = None
    tenant_id = None
    event = None
    event_by_uid = None
    instances_event = []
    extension_event = None

    def _get_access_token(self):
        return self.token

    def applications(self):
        return None

    def subscriptions(self):
        return self._subscriptions

    def subscription(self, subscription_id: str):
        return self._subscription

    def _renew_subscription(self, subscription_id: str):
        return self._subscription

    def get_tenant_id(self, email: str):
        return self.tenant_id

    def _subscribe(self, resource: str, change_type: str):
        return self._subscription

    def _unsubscribe(self, subscription_id: str):
        return None

    def get_event_by_resource(self, resource: str):
        return self.event

    def get_event(self, tenant_id: str, external_id: str):
        return self.event

    def get_event_by_uid(self, tenant_id: str, uid: str):
        return self.event_by_uid

    def get_instances_event(self, tenant_id: str, external_id: str, start: str, end: str):
        return self.instances_event

    def get_extension_event(self, tenant_id: str, extension_id: str):
        return self.extension_event


@pytest.mark.django_db
class TestOutlookSyncFixture:
    @pytest.fixture()
    def fixture_request(self):
        request = APIRequestFactory().get("")
        request.session = {}  # for sessions middleware
        request._messages = storage.default_storage(request)  # for messages middleware
        return request

    @pytest.fixture()
    def credentials_fixture(self):
        return '{"notification_url": "https://test.com/api/crm/sync/activity/event_watch", "authority": "https://test.com/fake_app_tenant_id", "client_id": "fake_client_id", "client_secret": "fake_client_secret", "token_endpoint": "/fake_token", "graph_url": "https://fake_url"}'

    @pytest.fixture()
    def notification_created_fixture(self):
        return {
            "subscription_id": "fake_subscription_id",
            "subscription_expiration": "2022-12-11T09:05:13+00:00",
            "change_type": "created",
            "resource": "Users/fake_user_id/Events/fake_event_id",
            "resource_data": {
                "@odata.etag": "fake_etag",
                "@odata.id": "Users/fake_user_id/Events/fake_event_id",
                "@odata.type": "#Microsoft.Graph.Event",
                "id": "unique_id",
            },
            "client_state": "secretClientValue",
            "tenant_id": "fake_tenant_id",
        }

    @pytest.fixture()
    def notification_fixture(self, notification_created_fixture):
        _fixture = copy.deepcopy(notification_created_fixture)
        _fixture["change_type"] = "deleted"
        return _fixture

    @pytest.fixture()
    def notification_call_record_fixture(self, notification_created_fixture):
        notification_call_record_fixture = copy.deepcopy(notification_created_fixture)
        notification_call_record_fixture["resource_data"]["@odata.type"] = "/communications/callRecords"
        return notification_call_record_fixture

    @pytest.fixture()
    def organizer_event_fixture(self):
        return {
            "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users('cd209b0b-3f83-4c35-82d2-d88a61820480')/events/$entity",
            "@odata.etag": 'W/"ZlnW4RIAV06KYYwlrfNZvQAALfZeRQ=="',
            "id": "AAMkAGI1AAAt9AHjAAA=",
            "createdDateTime": "2017-04-15T03:00:50.7579581Z",
            "lastModifiedDateTime": "2017-04-15T03:00:51.245372Z",
            "changeKey": "ZlnW4RIAV06KYYwlrfNZvQAALfZeRQ==",
            "categories": [],
            "originalStartTimeZone": "Pacific Standard Time",
            "originalEndTimeZone": "Pacific Standard Time",
            "iCalUId": "040000008200E00074C5B7101A82E00800000000DA2B357D94B5D201000000000000000010000000EC4597557F0CB34EA4CC2887EA7B17C3",
            "reminderMinutesBeforeStart": 15,
            "isReminderOn": True,
            "hasAttachments": False,
            "hideAttendees": False,
            "subject": "Let's go brunch",
            "bodyPreview": "Does noon work for you?",
            "importance": "normal",
            "sensitivity": "normal",
            "isAllDay": False,
            "isCancelled": False,
            "isDraft": False,
            "isOrganizer": True,
            "responseRequested": True,
            "seriesMasterId": None,
            "transactionId": "7E163156-7762-4BEB-A1C6-729EA81755A7",
            "showAs": "busy",
            "type": "singleInstance",
            "webLink": "https://outlook.office365.com/owa/?itemid=AAMkAGI1AAAt9AHjAAA%3D&exvsurl=1&path=/calendar/item",
            "onlineMeetingUrl": None,
            "isOnlineMeeting": True,
            "onlineMeetingProvider": "unknown",
            "onlineMeeting": None,
            "allowNewTimeProposals": True,
            "responseStatus": {"response": "organizer", "time": "0001-01-01T00:00:00Z"},
            "body": {
                "contentType": "html",
                "content": "<html><head></head><body>Does late morning work for you?</body></html>",
            },
            "start": {"dateTime": "2017-04-15T11:00:00.0000000", "timeZone": "Pacific Standard Time"},
            "end": {"dateTime": "2017-04-15T12:00:00.0000000", "timeZone": "Pacific Standard Time"},
            "location": {
                "displayName": "Harry's Bar",
                "locationType": "default",
                "uniqueId": "Harry's Bar",
                "uniqueIdType": "private",
            },
            "locations": [{"displayName": "Harry's Bar", "locationType": "default", "uniqueIdType": "unknown"}],
            "recurrence": None,
            "attendees": [
                {
                    "type": "required",
                    "status": {"response": "none", "time": "0001-01-01T00:00:00Z"},
                    "emailAddress": {"name": "Samantha Booth", "address": "samanthab@contoso.onmicrosoft.com"},
                }
            ],
            "organizer": {"emailAddress": {"name": "Dana Swope", "address": "danas@contoso.onmicrosoft.com"}},
        }

    @pytest.fixture()
    def invitation_event_fixture(self, organizer_event_fixture):
        invitation_event_fixture = copy.deepcopy(organizer_event_fixture)
        invitation_event_fixture["isOrganizer"] = False
        return invitation_event_fixture

    @pytest.fixture()
    def organizer_event_fixture_parsed(self):
        return {
            "odata_context": "https://graph.microsoft.com/v1.0/$metadata#users('cd209b0b-3f83-4c35-82d2-d88a61820480')/events/$entity",
            "odata_etag": 'W/"ZlnW4RIAV06KYYwlrfNZvQAALfZeRQ=="',
            "id": "AAMkAGI1AAAt9AHjAAA=",
            "created_date_time": "2017-04-15T03:00:50.7579581Z",
            "last_modified_date_time": "2017-04-15T03:00:51.245372Z",
            "change_key": "ZlnW4RIAV06KYYwlrfNZvQAALfZeRQ==",
            "categories": [],
            "original_start_time_zone": "Pacific Standard Time",
            "original_end_time_zone": "Pacific Standard Time",
            "uid": "040000008200E00074C5B7101A82E00800000000DA2B357D94B5D201000000000000000010000000EC4597557F0CB34EA4CC2887EA7B17C3",
            "reminder_minutes_before_start": 15,
            "is_reminder_on": True,
            "has_attachments": False,
            "hide_attendees": False,
            "subject": "Let's go brunch",
            "body_preview": "Does noon work for you?",
            "importance": "normal",
            "sensitivity": "normal",
            "is_all_day": False,
            "is_cancelled": False,
            "is_draft": False,
            "is_organizer": True,
            "response_requested": True,
            "series_master_id": None,
            "transaction_id": "7E163156-7762-4BEB-A1C6-729EA81755A7",
            "show_as": "busy",
            "type": "singleInstance",
            "web_link": "https://outlook.office365.com/owa/?itemid=AAMkAGI1AAAt9AHjAAA%3D&exvsurl=1&path=/calendar/item",
            "online_meeting_url": None,
            "is_online_meeting": True,
            "online_meeting_provider": "unknown",
            "online_meeting": None,
            "allow_new_time_proposals": True,
            "locations": [{"displayName": "Harry's Bar", "locationType": "default", "uniqueIdType": "unknown"}],
            "recurrence": None,
            "attendees": [
                {
                    "type": "required",
                    "status": {"response": "none", "time": "0001-01-01T00:00:00Z"},
                    "emailAddress": {"name": "Samantha Booth", "address": "samanthab@contoso.onmicrosoft.com"},
                }
            ],
            "response_status.response": "organizer",
            "response_status.time": "0001-01-01T00:00:00Z",
            "body.content_type": "html",
            "body.content": "<html><head></head><body>Does late morning work for you?</body></html>",
            "start.date_time": "2017-04-15T11:00:00.0000000",
            "start.time_zone": "Pacific Standard Time",
            "end.date_time": "2017-04-15T12:00:00.0000000",
            "end.time_zone": "Pacific Standard Time",
            "location.display_name": "Harry's Bar",
            "location.location_type": "default",
            "location.unique_id": "Harry's Bar",
            "location.unique_id_type": "private",
            "organizer.email_address.name": "Dana Swope",
            "organizer.email_address.address": "danas@contoso.onmicrosoft.com",
        }

    @pytest.fixture()
    def invitation_event_fixture_parsed(self, organizer_event_fixture_parsed):
        invitation_event_fixture_parsed = copy.deepcopy(organizer_event_fixture_parsed)
        invitation_event_fixture_parsed["is_organizer"] = False
        return invitation_event_fixture_parsed

    @pytest.fixture()
    def organizer_master_event_fixture_parsed(self):
        return {
            "id": "unique_id2-ocurrence_pattern_id-unique_recurrence_id2",
            "uid": "unique_fake_id",
            "type": "seriesMaster",
            "start.date_time": "2022-12-12T07:00:00.0000000",
            "end.date_time": "2022-12-12T07:30:00.0000000",
            "created": "2022-12-08T10:37:29.8938372Z",
            "show_as": "busy",
            "subject": "test recc",
            "is_draft": False,
            "web_link": "fake_web_link2",
            "locations": [],
            "categories": [],
            "change_key": "fake_change_key2",
            "importance": "normal",
            "is_all_day": False,
            "odata_etag": "fake_etag2",
            "sensitivity": "normal",
            "body.content": "",
            "body_preview": "",
            "is_cancelled": False,
            "is_organizer": True,
            "participants": [],
            "end.time_zone": "UTC",
            "last_modified": "2022-12-08T10:37:30.0188091Z",
            "occurrence_id": None,
            "odata_context": "fake_odata_context2",
            "hide_attendees": False,
            "is_reminder_on": True,
            "online_meeting": None,
            "transaction_id": "70957719-1fb2-5642-7bb6-c1363e1091e0",
            "has_attachments": False,
            "start.time_zone": "UTC",
            "series_master_id": None,
            "body.content_type": "html",
            "is_online_meeting": False,
            "online_meeting_url": None,
            "response_requested": True,
            "response_status.time": "0001-01-01T00:00:00Z",
            "location.address.type": "unknown",
            "location.display_name": "",
            "recurrence.range.type": "endDate",
            "location.location_type": "default",
            "original_end_time_zone": "W. Europe Standard Time",
            "location.unique_id_type": "unknown",
            "online_meeting_provider": "unknown",
            "recurrence.pattern.type": "daily",
            "allow_new_time_proposals": True,
            "original_start_time_zone": "W. Europe Standard Time",
            "recurrence.pattern.index": "first",
            "recurrence.pattern.month": 0,
            "response_status.response": "organizer",
            "recurrence.range.end_date": "2022-12-14",
            "recurrence.pattern.interval": 1,
            "recurrence.range.start_date": "2022-12-12",
            "organizer.email_address.name": "Fake organizer name",
            "calendar@odata.navigationLink": "fake_navigation_link",
            "reminder_minutes_before_start": 15,
            "calendar@odata.associationLink": "fake_association_link",
            "organizer.email_address.address": "test@test.com",
            "recurrence.pattern.day_of_month": 0,
            "recurrence.pattern.first_day_of_week": "sunday",
            "recurrence.range.recurrence_time_zone": "W. Europe Standard Time",
            "recurrence.range.number_of_occurrences": 0,
        }

    @pytest.fixture()
    def teams_event_fixture(self):
        return {
            "odata_etag": 'W/"AAAAAAAAAAAAAAAAA=="',
            "id": "eventid1",
            "created_date_time": "2023-12-14T10:23:43.4384008Z",
            "last_modified_date_time": "2023-12-14T10:25:13.7822553Z",
            "change_key": "AAAAAAAAAAAAAAAAA==",
            "categories": [],
            "transaction_id": None,
            "original_start_time_zone": "UTC",
            "original_end_time_zone": "UTC",
            "uid": "eventuid1",
            "reminder_minutes_before_start": 15,
            "is_reminder_on": True,
            "has_attachments": False,
            "subject": "Canceled: Weekly Office meeting",
            "body_preview": "________________________________________________________________________________\r\nRéunion Microsoft Teams\r\nParticipez à partir de votre ordinateur, de votre application mobile ou de l’appareil de la salle\r\nCliquez ici pour rejoindre la réunion\r\nID de la",
            "importance": "normal",
            "sensitivity": "normal",
            "is_all_day": False,
            "is_cancelled": False,
            "is_organizer": False,
            "response_requested": True,
            "series_master_id": None,
            "show_as": "free",
            "type": "singleInstance",
            "web_link": "https://test.ch",
            "online_meeting_url": None,
            "is_online_meeting": True,
            "online_meeting_provider": "teamsForBusiness",
            "allow_new_time_proposals": True,
            "occurrence_id": None,
            "is_draft": False,
            "hide_attendees": False,
            "locations": [
                {
                    "displayName": "Microsoft Teams Meeting",
                    "locationType": "default",
                    "uniqueId": "Microsoft Teams Meeting",
                    "uniqueIdType": "private",
                }
            ],
            "recurrence": None,
            "attendees": [
                {
                    "type": "required",
                    "status": {"response": "none", "time": "0001-01-01T00:00:00Z"},
                    "emailAddress": {"name": "organizer1", "address": "organizer1@test.ch"},
                },
                *[
                    {
                        "type": "optional",
                        "status": {"response": "none", "time": "0001-01-01T00:00:00Z"},
                        "emailAddress": {"name": f"attendee{i}", "address": f"attendee{i}@atonra.ch"},
                    }
                    for i in range(1, 15)
                ],
            ],
            "response_status.response": "notResponded",
            "response_status.time": "0001-01-01T00:00:00Z",
            "body.content_type": "html",
            "start.date_time": "2023-12-18T08:30:00.0000000",
            "start.time_zone": "UTC",
            "end.date_time": "2023-12-18T10:00:00.0000000",
            "end.time_zone": "UTC",
            "location.display_name": "Microsoft Teams Meeting",
            "location.location_type": "default",
            "location.unique_id": "Microsoft Teams Meeting",
            "location.unique_id_type": "private",
            "organizer.email_address.name": "organizer1",
            "organizer.email_address.address": "organizer1@atonra.ch",
        }

    @pytest.fixture()
    def canceled_teams_event_fixture(self, teams_event_fixture):
        canceled_event = copy.deepcopy(teams_event_fixture)
        canceled_event["is_cancelled"] = True
        return canceled_event


# recurrence_event = {
#     "id": "unique_id2-ocurrence_pattern_id-unique_recurrence_id2",
#     "uid": event.get("uid"),
#     "type": "seriesMaster",
#     "start": "2022-12-12T07:00:00.0000000",
#     "end": "2022-12-12T07:30:00.0000000",
#     "created": "2022-12-08T10:37:29.8938372Z",
#     "show_as": "busy",
#     "subject": "test recc",
#     "is_draft": False,
#     "web_link": "fake_web_link2",
#     "locations": [],
#     "categories": [],
#     "change_key": "fake_change_key2",
#     "importance": "normal",
#     "is_all_day": False,
#     "odata_etag": "fake_etag2",
#     "sensitivity": "normal",
#     "body.content": "",
#     "body_preview": "",
#     "is_cancelled": False,
#     "is_organizer": True,
#     "participants": [],
#     "end.time_zone": "UTC",
#     "last_modified": "2022-12-08T10:37:30.0188091Z",
#     "occurrence_id": None,
#     "odata_context": "fake_odata_context2",
#     "hide_attendees": False,
#     "is_reminder_on": True,
#     "online_meeting": None,
#     "transaction_id": "70957719-1fb2-5642-7bb6-c1363e1091e0",
#     "has_attachments": False,
#     "start.time_zone": "UTC",
#     "series_master_id": None,
#     "body.content_type": "html",
#     "is_online_meeting": False,
#     "online_meeting_url": None,
#     "response_requested": True,
#     "response_status.time": "0001-01-01T00:00:00Z",
#     "location.address.type": "unknown",
#     "location.display_name": "",
#     "recurrence.range.type": "endDate",
#     "location.location_type": "default",
#     "original_end_time_zone": "W. Europe Standard Time",
#     "location.unique_id_type": "unknown",
#     "online_meeting_provider": "unknown",
#     "recurrence.pattern.type": "daily",
#     "allow_new_time_proposals": True,
#     "original_start_time_zone": "W. Europe Standard Time",
#     "recurrence.pattern.index": "first",
#     "recurrence.pattern.month": 0,
#     "response_status.response": "organizer",
#     "recurrence.range.end_date": "2022-12-14",
#     "recurrence.pattern.interval": 1,
#     "recurrence.range.start_date": "2022-12-12",
#     "organizer.email_address.name": "Fake organizer name",
#     "calendar@odata.navigationLink": "fake_navigation_link",
#     "reminder_minutes_before_start": 15,
#     "calendar@odata.associationLink": "fake_association_link",
#     "organizer.email_address.address": "test@test.com",
#     "recurrence.pattern.day_of_month": 0,
#     "recurrence.pattern.first_day_of_week": "sunday",
#     "recurrence.range.recurrence_time_zone": "W. Europe Standard Time",
#     "recurrence.range.number_of_occurrences": 0,
# }

# occurrences_event = [
#     {
#         "id": "unique_id001-ocurrence_pattern_id-unique_recurrence_id2",
#         "end": "2022-12-12T07:30:00.0000000",
#         "uid": event.get("uid"),
#         "type": "occurrence",
#         "start": "2022-12-12T07:00:00.0000000",
#         "created": "2022-12-08T10:37:29.8938372Z",
#         "show_as": "busy",
#         "subject": "test recc",
#         "is_draft": False,
#         "web_link": "weblink",
#         "locations": [],
#         "categories": [],
#         "change_key": "changekey1",
#         "importance": "normal",
#         "is_all_day": False,
#         "odata_etag": "etag1",
#         "recurrence": None,
#         "sensitivity": "normal",
#         "body.content": "",
#         "body_preview": "",
#         "is_cancelled": False,
#         "is_organizer": True,
#         "participants": [],
#         "end.time_zone": "UTC",
#         "last_modified": "2022-12-08T10:37:30.0188091Z",
#         "occurrence_id": "OID.unique_id2-ocurrence_pattern_id-unique_recurrence_id2.2022-12-12",
#         "hide_attendees": False,
#         "is_reminder_on": True,
#         "online_meeting": None,
#         "transaction_id": "fake_transaction_id",
#         "has_attachments": False,
#         "start.time_zone": "UTC",
#         "series_master_id": "unique_id2-ocurrence_pattern_id-unique_recurrence_id2",
#         "body.content_type": "html",
#         "is_online_meeting": False,
#         "online_meeting_url": None,
#         "response_requested": True,
#         "response_status.time": "0001-01-01T00:00:00Z",
#         "location.address.type": "unknown",
#         "location.display_name": "",
#         "location.location_type": "default",
#         "original_end_time_zone": "W. Europe Standard Time",
#         "location.unique_id_type": "unknown",
#         "online_meeting_provider": "unknown",
#         "allow_new_time_proposals": True,
#         "original_start_time_zone": "W. Europe Standard Time",
#         "response_status.response": "organizer",
#         "organizer.email_address.name": "Fake organizer name",
#         "reminder_minutes_before_start": 15,
#         "organizer.email_address.address": "test@test.com",
#     },
#     {
#         "id": "unique_id002-ocurrence_pattern_id-unique_recurrence_id2",
#         "end": "2022-12-13T07:30:00.0000000",
#         "uid": event.get("uid"),
#         "type": "occurrence",
#         "start": "2022-12-13T07:00:00.0000000",
#         "created": "2022-12-08T10:37:29.8938372Z",
#         "show_as": "busy",
#         "subject": "test recc",
#         "is_draft": False,
#         "web_link": "weblink",
#         "locations": [],
#         "categories": [],
#         "change_key": "changekey1",
#         "importance": "normal",
#         "is_all_day": False,
#         "odata_etag": "etag1",
#         "recurrence": None,
#         "sensitivity": "normal",
#         "body.content": "",
#         "body_preview": "",
#         "is_cancelled": False,
#         "is_organizer": True,
#         "participants": [],
#         "end.time_zone": "UTC",
#         "last_modified": "2022-12-08T10:37:30.0188091Z",
#         "occurrence_id": "OID.unique_id2-ocurrence_pattern_id-unique_recurrence_id2.2022-12-13",
#         "hide_attendees": False,
#         "is_reminder_on": True,
#         "online_meeting": None,
#         "transaction_id": "transaction1",
#         "has_attachments": False,
#         "start.time_zone": "UTC",
#         "series_master_id": "unique_id2-ocurrence_pattern_id-unique_recurrence_id2",
#         "body.content_type": "html",
#         "is_online_meeting": False,
#         "online_meeting_url": None,
#         "response_requested": True,
#         "response_status.time": "0001-01-01T00:00:00Z",
#         "location.address.type": "unknown",
#         "location.display_name": "",
#         "location.location_type": "default",
#         "original_end_time_zone": "W. Europe Standard Time",
#         "location.unique_id_type": "unknown",
#         "online_meeting_provider": "unknown",
#         "allow_new_time_proposals": True,
#         "original_start_time_zone": "W. Europe Standard Time",
#         "response_status.response": "organizer",
#         "organizer.email_address.name": "Fake organizer name",
#         "reminder_minutes_before_start": 15,
#         "organizer.email_address.address": "test@test.com",
#     },
#     {
#         "id": "unique_id003-ocurrence_pattern_id-unique_recurrence_id2",
#         "end": "2022-12-14T07:30:00.0000000",
#         "uid": event.get("uid"),
#         "type": "occurrence",
#         "start": "2022-12-14T07:00:00.0000000",
#         "created": "2022-12-08T10:37:29.8938372Z",
#         "show_as": "busy",
#         "subject": "test recc",
#         "is_draft": False,
#         "web_link": "weblink4",
#         "locations": [],
#         "categories": [],
#         "change_key": "change_key5",
#         "importance": "normal",
#         "is_all_day": False,
#         "odata_etag": "etag3",
#         "recurrence": None,
#         "sensitivity": "normal",
#         "body.content": "",
#         "body_preview": "",
#         "is_cancelled": False,
#         "is_organizer": True,
#         "participants": [],
#         "end.time_zone": "UTC",
#         "last_modified": "2022-12-08T10:37:30.0188091Z",
#         "occurrence_id": "OID.unique_id2-ocurrence_pattern_id-unique_recurrence_id2..2022-12-14",
#         "hide_attendees": False,
#         "is_reminder_on": True,
#         "online_meeting": None,
#         "transaction_id": "transaction3",
#         "has_attachments": False,
#         "start.time_zone": "UTC",
#         "series_master_id": "unique_id2-ocurrence_pattern_id-unique_recurrence_id2",
#         "body.content_type": "html",
#         "is_online_meeting": False,
#         "online_meeting_url": None,
#         "response_requested": True,
#         "response_status.time": "0001-01-01T00:00:00Z",
#         "location.address.type": "unknown",
#         "location.display_name": "",
#         "location.location_type": "default",
#         "original_end_time_zone": "W. Europe Standard Time",
#         "location.unique_id_type": "unknown",
#         "online_meeting_provider": "unknown",
#         "allow_new_time_proposals": True,
#         "original_start_time_zone": "W. Europe Standard Time",
#         "response_status.response": "organizer",
#         "organizer.email_address.name": "Fake organizer name",
#         "reminder_minutes_before_start": 15,
#         "organizer.email_address.address": "test@test.com",
#     },
# ]



from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Company

from wbcrm.models import Product, ProductCompanyRelationship


class ProductFilterSet(wb_filters.FilterSet):
    prospects = wb_filters.ModelMultipleChoiceFilter(
        label=_("Prospects"),
        queryset=Company.objects.all(),
        endpoint=Company.get_representation_endpoint(),
        value_key=Company.get_representation_value_key(),
        label_key=Company.get_representation_label_key(),
    )

    class Meta:
        model = Product
        fields = {
            "title": ["exact", "icontains"],
            "is_competitor": ["exact"],
        }


class ProductCompanyFilterSet(wb_filters.FilterSet):
    competitor_product = wb_filters.BooleanFilter(label=_("Is Competitor"), method="filter_competitor_product")

    def filter_competitor_product(self, queryset, label, value):
        if value is None:
            return queryset
        else:
            return queryset.filter(product__is_competitor=value)

    class Meta:
        model = ProductCompanyRelationship
        fields = {
            "product": ["exact"],
        }



from django.db.models import Case, Exists, F, IntegerField, OuterRef, When
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from rest_framework.exceptions import AuthenticationFailed
from wbcore import viewsets
from wbcore.utils.views import MergeMixin

from wbcrm.filters import AccountFilter, AccountRoleFilterSet
from wbcrm.models.accounts import Account, AccountRole, AccountRoleType, AccountRoleValidity
from wbcrm.serializers.accounts import (
    AccountModelSerializer,
    AccountRepresentationSerializer,
    AccountRoleModelSerializer,
    AccountRoleTypeRepresentationSerializer,
    AccountRoleValidityModelSerializer,
)
from wbcrm.viewsets.buttons import AccountButtonConfig
from wbcrm.viewsets.display import (
    AccountDisplayConfig,
    AccountRoleAccountDisplayConfig,
    AccountRoleValidityDisplayConfig,
    InheritedAccountRoleAccountDisplayConfig,
)
from wbcrm.viewsets.endpoints import (
    AccountRoleAccountEndpointConfig,
    AccountRoleValidityEndpointConfig,
    ChildAccountAccountEndpointConfig,
    InheritedAccountRoleAccountEndpointConfig,
)
from wbcrm.viewsets.titles import (
    AccountRoleAccountTitleConfig,
    AccountTitleConfig,
    ChildAccountAccountTitleConfig,
)

from .mixins import AccountPermissionMixin, AccountRolePermissionMixin


class AccountRoleTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = search_fields = ordering = ("title",)
    serializer_class = AccountRoleTypeRepresentationSerializer
    queryset = AccountRoleType.objects.all()


class AccountRepresentationViewSet(AccountPermissionMixin, viewsets.RepresentationViewSet):
    ordering_fields = ("title",)
    search_fields = ["computed_str", "owner__computed_str"]
    filterset_class = AccountFilter

    serializer_class = AccountRepresentationSerializer
    queryset = Account.all_objects.all()
    ordering = ["title"]


class AccountModelViewSet(MergeMixin, AccountPermissionMixin, viewsets.ModelViewSet):
    serializer_class = AccountModelSerializer
    filterset_class = AccountFilter
    queryset = Account.all_objects.select_related("owner", "parent").annotate(
        has_children=Exists(Account.objects.filter(parent=OuterRef("pk"))),
        _group_key=Case(When(has_children=True, then=F("id")), default=None, output_field=IntegerField()),
    )

    ordering = ordering_fields = [
        "title",
        "owner__computed_str",
        "reference_id",
        "is_terminal_account",
        "is_public",
        "is_active",
    ]
    search_fields = ["computed_str", "owner__computed_str"]

    display_config_class = AccountDisplayConfig
    title_config_class = AccountTitleConfig
    button_config_class = AccountButtonConfig

    def get_merged_object_representation_serializer(self):
        return AccountRepresentationSerializer


class ChildAccountAccountModelViewSet(AccountModelViewSet):
    title_config_class = ChildAccountAccountTitleConfig
    endpoint_config_class = ChildAccountAccountEndpointConfig

    def dispatch(self, *args, **kwargs):
        kwargs["parent_id"] = kwargs.get("account_id", None)
        return super().dispatch(*args, **kwargs)

    def get_queryset(self):
        return super().get_queryset().filter(parent_id=self.kwargs["account_id"])


class AccountRoleAccountModelViewSet(AccountRolePermissionMixin, viewsets.ModelViewSet):
    display_config_class = AccountRoleAccountDisplayConfig
    title_config_class = AccountRoleAccountTitleConfig
    endpoint_config_class = AccountRoleAccountEndpointConfig
    serializer_class = AccountRoleModelSerializer

    filterset_class = AccountRoleFilterSet
    queryset = AccountRole.objects.select_related(
        "entry",
        "account",
    )
    ordering = ["role_type", "id"]

    @cached_property
    def account(self):
        account = get_object_or_404(Account, pk=self.kwargs["account_id"])
        if not account.can_see_account(self.request.user):
            raise AuthenticationFailed()
        return account

    def get_queryset(self):
        return super().get_queryset().filter(account=self.account)


class InheritedAccountRoleAccountModelViewSet(AccountRoleAccountModelViewSet):
    display_config_class = InheritedAccountRoleAccountDisplayConfig
    endpoint_config_class = InheritedAccountRoleAccountEndpointConfig
    ONLY_READ_ONLY_ENDPOINT = True

    def get_queryset(self):
        return AccountRole.objects.filter(
            id__in=self.account.get_inherited_roles_for_account().values("id")
        ).filter_for_user(self.request.user, validity_date=self.validity_date)


class AccountRoleValidityModelViewSet(viewsets.ModelViewSet):
    serializer_class = AccountRoleValidityModelSerializer
    queryset = AccountRoleValidity.objects.all()
    display_config_class = AccountRoleValidityDisplayConfig
    endpoint_config_class = AccountRoleValidityEndpointConfig

    @cached_property
    def role(self):
        return get_object_or_404(AccountRole, pk=self.kwargs["account_role_id"])

    def get_queryset(self):
        return super().get_queryset().filter(role=self.role)



from wbcore.metadata.configs.titles import TitleViewConfig

from wbcrm.models.accounts import Account


class AccountTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Account: {{title}}"

    def get_list_title(self):
        return "Accounts"


class ChildAccountAccountTitleConfig(TitleViewConfig):
    def get_list_title(self):
        account = Account.all_objects.get(id=self.view.kwargs["account_id"])
        return f"{account.title}: Sub Accounts"


class AccountRoleAccountTitleConfig(TitleViewConfig):
    def get_list_title(self):
        account = Account.all_objects.get(id=self.view.kwargs["account_id"])
        return f"Roles for Account {str(account)}"



# Generated by Django 5.2.9 on 2025-12-16 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0014_alter_entry_relationship_managers_and_more'),
        ('wbcrm', '0019_alter_activity_companies_alter_activity_participants_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='product',
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name='product',
            constraint=models.UniqueConstraint(fields=('slugify_title', 'is_competitor'), name='unique_product'),
        ),
    ]



from django.apps import apps
from django.contrib.messages import warning
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from rest_framework.validators import UniqueValidator
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.models import User
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer

from wbcrm.models.accounts import Account, AccountRole, AccountRoleType, AccountRoleValidity


class AccountRoleTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = AccountRoleType
        fields = ("id", "title")


class AccountRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcrm:account-detail")

    def get_filter_params(self, request):
        return {"is_active": True}

    class Meta:
        model = Account
        fields = ("id", "computed_str", "_detail")


class TerminalAccountRepresentationSerializer(AccountRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = {"is_terminal_account": True, "status": "OPEN", "is_active": True}
        if view := request.parser_context.get("view", None):
            if entry_id := view.kwargs.get("entry_id", None):
                filter_params["customer"] = entry_id
            if account_id := view.kwargs.get("account_id", None):
                filter_params["account"] = account_id
        return filter_params


class AccountModelSerializer(wb_serializers.ModelSerializer):
    _parent = AccountRepresentationSerializer(source="parent")
    _owner = EntryRepresentationSerializer(source="owner")
    _group_key = wb_serializers.CharField(read_only=True)
    reference_id = wb_serializers.IntegerField(
        label="Reference ID",
        default=lambda: Account.get_next_available_reference_id(),
        read_only=lambda view: not view.new_mode,
        validators=[UniqueValidator(queryset=Account.objects.all())],
    )

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        request = self.context["request"]

        custom_buttons = {
            "childaccounts": reverse("wbcrm:account-childaccount-list", args=[instance.id], request=request),
            "accountroles": reverse("wbcrm:account-accountrole-list", args=[instance.id], request=request),
            "inheritedaccountroles": reverse("wbcrm:account-inheritedrole-list", args=[instance.id], request=request),
        }
        if apps.is_installed("wbportfolio"):  # TODO move as signal
            custom_buttons["claims"] = reverse("wbportfolio:account-claim-list", args=[instance.id], request=request)

        return custom_buttons

    relationship_status = wb_serializers.RangeSelectField(
        color="rgb(220,20,60)",
        label="Relationship Status",
        start=1,
        end=5,
        step_size=1,
        read_only=True,
        required=False,
    )

    class Meta:
        model = Account
        fields = (
            "id",
            "_group_key",
            "title",
            "status",
            "is_active",
            "is_terminal_account",
            "is_public",
            "_parent",
            "parent",
            "_owner",
            "owner",
            "reference_id",
            "computed_str",
            "relationship_status",
            "relationship_summary",
            "action_plan",
            "_additional_resources",
        )


class AccountRoleModelSerializer(wb_serializers.ModelSerializer):
    _entry = EntryRepresentationSerializer(source="entry")
    _account = AccountRepresentationSerializer(source="account")
    _role_type = AccountRoleTypeRepresentationSerializer(source="role_type")
    is_currently_valid = wb_serializers.BooleanField(read_only=True, default=False)

    _authorized_hidden_users = UserRepresentationSerializer(source="authorized_hidden_users", many=True)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "validity": reverse("wbcrm:accountrole-validity-list", args=[instance.id]),
        }

    def create(self, validated_data):
        # We return a get or create role because we don't want to leak information on already existing role that the user might not be able to see
        # this way, the account role (even if already existing) will be returned and the permission mixin will take over
        authorized_hidden_users = validated_data.pop("authorized_hidden_users", [])
        instance, created = AccountRole.objects.get_or_create(
            entry=validated_data.pop("entry"), account=validated_data.pop("account"), defaults=validated_data
        )
        if authorized_hidden_users:
            instance.authorized_hidden_users.set(authorized_hidden_users)
        if request := self.context.get("request"):
            if not User.objects.filter(profile_id=instance.entry.id).exists():
                warning(
                    request,
                    _(
                        "The selected entry does not have an associated user account. Note: Notifications cannot be sent to users without an account."
                    ),
                )
        return instance

    class Meta:
        model = AccountRole
        fields = (
            "id",
            "role_type",
            "_role_type",
            "entry",
            "_entry",
            "account",
            "_account",
            "is_currently_valid",
            "is_hidden",
            "authorized_hidden_users",
            "_authorized_hidden_users",
            "_additional_resources",
        )


class AccountRoleValidityModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = AccountRoleValidity
        fields = ("id", "timespan", "role")



from datetime import timedelta

from django.utils import timezone
from dynamic_preferences.registries import global_preferences_registry
from googleapiclient.discovery import Resource
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.directory.models import Person
from wbcrm.models import Activity, ActivityParticipant

from ...typing_informations import GoogleEventType
from ...utils import GoogleSyncUtils


def update_activity_participant(event: GoogleEventType, activity: Activity):
    """
    Used to update the participants in a workbench activity.

    :param event:   The Google event dict with the participant informations.
    :param activity:The corresponding activity.
    """

    can_sync_external_participants: bool = global_preferences_registry.manager()[
        "wbactivity_sync__sync_external_participants"
    ]
    event_participants = GoogleSyncUtils.get_or_create_participants(event, activity.creator)

    def update_or_add_participants():
        for event_participant in event_participants:
            if person := Person.objects.filter(id=event_participant["person_id"]).first():
                activity.participants.add(person.id)
                ActivityParticipant.objects.filter(activity=activity, participant=person).update(
                    participation_status=event_participant["status"]
                )

    if can_sync_external_participants or not activity.creator.is_internal:  # type: ignore
        update_or_add_participants()
    else:
        internal_activity_participants_set = set(
            activity.participants.filter(id__in=Person.objects.filter_only_internal())
        )
        event_participants_set = set(Person.objects.filter(id__in=[x["person_id"] for x in event_participants]))
        missing_activity_participants = internal_activity_participants_set - event_participants_set
        update_or_add_participants()
        if missing_activity_participants:
            ActivityParticipant.objects.filter(
                participant__in=missing_activity_participants, activity=activity
            ).update(participation_status=ActivityParticipant.ParticipationStatus.CANCELLED)


def update_single_activity(event: GoogleEventType, activity: Activity, change_status=False):
    """
    Updates a single workbench activity based on changes done in a google event.

    :param event:           The google event dict that was updated.
    :param activity:        The corresponding workbench activity.
    :param change_status:   Information wheter the status of the event was updated or not.
    """
    if activity.external_id is None:
        return
    if activity.status == Activity.Status.REVIEWED or activity.status == Activity.Status.FINISHED:
        metadata = activity.metadata
        metadata["google_backend"] |= {"event": event} if not event.get("recurringEventId") else {"instance": event}
        Activity.objects.filter(id=activity.id).update(
            external_id=event["id"] if "_" in activity.external_id else activity.external_id,
            metadata=metadata,
        )
        return
    event_organizer = event.get("organizer", {})
    event_organizer_mail = event_organizer.get("email", "")
    event_organizer_displayed_name = event_organizer.get("displayName", "")
    organizer = GoogleSyncUtils.get_or_create_person(event_organizer_mail, event_organizer_displayed_name)
    event_start, event_end = GoogleSyncUtils.get_start_and_end(event)
    if event_start == event_end:
        event_end = event_end + timedelta(seconds=1)
    period = TimestamptzRange(event_start, event_end)  # type: ignore
    all_day: bool = True if event["start"].get("date") else False

    metadata = activity.metadata
    metadata["google_backend"] |= {"event": event} if not event.get("recurringEventId") else {"instance": event}
    Activity.objects.filter(id=activity.id).update(
        title=event.get("summary", "(No Subject)"),
        assigned_to=organizer,
        description=event.get("description", ""),
        start=event_start,
        status=Activity.Status.PLANNED if change_status else activity.status,
        end=event_end,
        period=period,
        all_day=all_day,
        location=event.get("location"),
        visibility=GoogleSyncUtils.convert_event_visibility_to_activity_visibility(event.get("visibility")),
        metadata=metadata,
        external_id=event["id"] if "_" in activity.external_id else activity.external_id,
    )
    update_activity_participant(event, activity)


def update_all_activities(activity: Activity, event: GoogleEventType, user_mail: str, service: Resource):
    """
    Updates all workbench activities in a recurrence chain based on changes done in a google event.

    :param event:       The google event dict that was updated.
    :param activity:    The corresponding workbench activity.
    :param user_mail:   The e-mail address of the current user.
    :param service:     The google service to interact with googles resources.
    """

    activity_instance = activity.metadata["google_backend"].get(
        "instance", activity.metadata["google_backend"].get("event", {})
    )
    event_start, event_end = event["start"], event["end"]
    activity_start, activity_end = activity_instance.get("start"), activity_instance.get("end")
    event_instances: dict = service.events().instances(calendarId=user_mail, eventId=event["id"]).execute()
    instance_items = event_instances["items"]
    connected_activities = Activity.objects.filter(parent_occurrence=activity)
    connected_activities |= Activity.objects.filter(metadata__google_backend__instance__recurringEventId=event["id"])
    connected_activities |= Activity.objects.filter(id=activity.id)
    if event_start != activity_start or event_end != activity_end:
        # If the start/end time of the google event chain changes, google will also change the id of the event instances. That is why we need to handle this case seperatly
        connected_activities = connected_activities.order_by("external_id")
        activity_list = list(connected_activities)
        instance_items.sort(key=lambda x: x["start"].get("date", x["start"]["dateTime"]))
        for instance in instance_items:
            if len(activity_list) == 0:
                break
            activity_instance = activity_list.pop(0)
            update_single_activity(instance, activity_instance)
            activity_instance.refresh_from_db()
    else:
        for instance in instance_items:
            if activity_child := connected_activities.filter(external_id=instance["id"]).first() or (
                activity_child := Activity.objects.filter(
                    metadata__google_backend__instance__id=instance["id"]
                ).first()
            ):
                update_single_activity(instance, activity_child)


def update_activities_from_new_parent(event: dict, parent_occurrence: Activity, user_mail: str, service: Resource):
    """
    This methods updates child activities whose parent activity was altered.

    :param event:           The google event dict that was updated.
    :param parent_occurrence: The corresponding workbench parent activity.
    :param user_mail:       The e-mail address of the current user.
    :param service:         The google service to interact with googles resources.

    """
    now = timezone.now()
    canceled_child_activities = Activity.objects.filter(
        parent_occurrence=parent_occurrence, period__startswith__gt=now
    )
    event_instances: dict = service.events().instances(calendarId=user_mail, eventId=event["id"]).execute()
    for instance in event_instances["items"]:
        if activity := canceled_child_activities.filter(external_id=instance["id"]).first():
            update_single_activity(instance, activity, True)



from datetime import timedelta

import pytest
from django.utils import timezone
from faker import Faker
from wbcore.contrib.directory.models import Company

from wbcrm.factories import ActivityTypeFactory
from wbcrm.models import Activity, ActivityType
from wbcrm.tasks import default_activity_heat_calculation

fake = Faker()


@pytest.mark.django_db
class TestSpecificTasks:
    pass


@pytest.mark.django_db
class TestActivityHeatCalculationTasks:
    @pytest.mark.parametrize("end", [fake.date_time()])
    def test_high_activity_heat(self, company_factory, activity_factory, end):
        end = timezone.make_aware(end)

        high_activity_company_a = company_factory(name="Company A", activity_heat=0.0)
        high_activity_company_b = company_factory(name="Company B", activity_heat=0.0)
        high_activity_company_c = company_factory(name="Company C", activity_heat=0.0)
        activity_factory(
            status=Activity.Status.REVIEWED,
            companies=[high_activity_company_a.id, high_activity_company_b.id],
            type=ActivityTypeFactory(score=ActivityType.Score.HIGH),
            start=end - timedelta(hours=2),
            end=end - timedelta(hours=1),
        )

        activity_factory(
            status=Activity.Status.REVIEWED,
            companies=[high_activity_company_b.id],
            type=ActivityTypeFactory(score=ActivityType.Score.LOW),
            start=end - timedelta(hours=2),
            end=end - timedelta(hours=1),
        )

        activity_factory(
            status=Activity.Status.REVIEWED,
            companies=[
                high_activity_company_c.id,
            ],
            type=ActivityTypeFactory(score=ActivityType.Score.HIGH),
            start=end - timedelta(days=1, hours=2),
            end=end - timedelta(days=1, hours=1),
        )

        default_activity_heat_calculation(end)
        high_activity_company_a: Company = Company.objects.get(id=high_activity_company_a.id)
        high_activity_company_b: Company = Company.objects.get(id=high_activity_company_b.id)
        high_activity_company_c: Company = Company.objects.get(id=high_activity_company_c.id)
        date_score = 365 / 365
        target_score_a = (float(ActivityType.Score.HIGH) * date_score) / float(ActivityType.Score.MAX)
        target_score_b = (
            (float(ActivityType.Score.HIGH) * date_score) + (float(ActivityType.Score.LOW) * date_score)
        ) / float(ActivityType.Score.MAX)
        target_score_c = (float(ActivityType.Score.HIGH) * (364 / 365)) / float(ActivityType.Score.MAX)
        assert round(high_activity_company_a.activity_heat, 1) == round(target_score_a, 1)
        assert round(high_activity_company_b.activity_heat, 1) == round(target_score_b, 1)
        assert round(high_activity_company_c.activity_heat, 1) == round(target_score_c, 1)

    @pytest.mark.parametrize("end", [fake.date_time()])
    def test_low_activity_heat(self, company_factory, activity_factory, end):
        end = timezone.make_aware(end)

        low_activity_company_a = company_factory(name="Company A", activity_heat=0.0)
        low_activity_company_b = company_factory(name="Company B", activity_heat=0.0)
        activity_factory(
            status=Activity.Status.REVIEWED,
            companies=[low_activity_company_a.id],
            type=ActivityTypeFactory(score=ActivityType.Score.LOW),
            start=end - timedelta(hours=2),
            end=end - timedelta(hours=1),
        )
        activity_factory(
            status=Activity.Status.REVIEWED,
            companies=[low_activity_company_b.id],
            type=ActivityTypeFactory(score=ActivityType.Score.HIGH),
            start=end - timedelta(hours=2),
            end=end - timedelta(hours=1),
        )

        default_activity_heat_calculation(end)
        low_activity_company_a = Company.objects.get(id=low_activity_company_a.id)
        date_score = (365 - 100) / 365
        target_score_a = (float(ActivityType.Score.LOW) * date_score) / float(ActivityType.Score.MAX)
        assert round(low_activity_company_a.activity_heat, 1) == round(target_score_a, 1)
        assert low_activity_company_b.activity_heat == 0.0



from .accounts import (
    AccountModelSerializer,
    AccountRepresentationSerializer,
    TerminalAccountRepresentationSerializer,
    AccountRoleModelSerializer,
    AccountRoleTypeRepresentationSerializer,
    AccountRoleValidityModelSerializer
)
from .activities import (
    ActivityModelListSerializer,
    ActivityModelSerializer,
    ReadOnlyActivityModelSerializer,
    ActivityParticipantModelSerializer,
    ActivityRepresentationSerializer,
    ActivityTypeModelSerializer,
    ActivityTypeRepresentationSerializer,
)
from .groups import GroupModelSerializer, GroupRepresentationSerializer
from .products import (
    ProductCompanyRelationshipModelSerializer,
    ProductModelSerializer,
    ProductRepresentationSerializer,
)
from .signals import *



from .accounts import (
    AccountDisplayConfig,
    AccountRoleAccountDisplayConfig,
    InheritedAccountRoleAccountDisplayConfig,
AccountRoleValidityDisplayConfig
)
from .activities import (
    ActivityDisplay,
    ActivityParticipantDisplayConfig,
    ActivityTypeDisplay,
)
from .groups import GroupModelDisplay
from .products import ProductCompanyRelationshipDisplay, ProductDisplay



from datetime import timedelta

from django.db.models.query import QuerySet
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from psycopg.types.range import TimestamptzRange
from wbcore import filters as wb_filters
from wbcore.contrib.directory.filters.entries import (
    CompanyFilter,
    EntryFilter,
    PersonFilter,
    UserIsManagerFilter,
)
from wbcore.contrib.directory.models import Entry
from wbcore.signals.filters import add_filters

from wbcrm.models import Account, AccountRole, Product


def choice_noactivity(queryset: QuerySet[Entry], name, value: str | None) -> QuerySet[Entry]:
    """Returns only those entries who didn't participate in any activities during the selected timeframe"""

    if value is None:
        return queryset
    else:
        value = int(value)
        if value:
            end = timezone.now()
            start = end - timedelta(days=value)
            no_activity_timeframe = TimestamptzRange(start, end)
            return queryset.exclude(
                calendar_entities__period__overlap=no_activity_timeframe,
                calendar_entities__item_type="wbcrm.Activity",
            )
        # A value of 0 corresponds to "All Time" in the filter so we need exclude any activities the entries participated in
        return queryset.exclude(calendar_entities__item_type="wbcrm.Activity")


@receiver(add_filters, sender=EntryFilter)
@receiver(add_filters, sender=PersonFilter)
@receiver(add_filters, sender=CompanyFilter)
@receiver(add_filters, sender=UserIsManagerFilter)
def add_account_filter(sender, request=None, *args, **kwargs):
    def _filter_with_account(queryset: QuerySet[Entry], name, value: str | None) -> QuerySet[Entry]:
        accounts = Account.objects.filter_for_user(request.user)
        if value:
            return queryset.filter(accounts__in=accounts).distinct()
        return queryset

    def _filter_with_account_role(queryset: QuerySet[Entry], name, value: str | None) -> QuerySet[Entry]:
        roles = AccountRole.objects.filter_for_user(request.user)
        if value:
            return queryset.filter(account_roles__in=roles).distinct()
        return queryset

    def filter_interested_products(queryset, name, value):
        if value:
            return queryset.filter(interested_products__in=value)
        return queryset

    interested_products = wb_filters.ModelMultipleChoiceFilter(
        label="Interested Products",
        help_text="Filter by products (ours and comptetitors) that the customer is interested in",
        queryset=Product.objects.all(),
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key=Product.get_representation_label_key(),
        method=filter_interested_products,
    )

    return {
        "without_activity": wb_filters.ChoiceFilter(
            label=_("No Activity"),
            help_text="Filter for entries without any activities",
            field_name="without_activity",
            choices=[
                (31, _("Last Month")),
                (92, _("Last 3 Months")),
                (182, _("Last 6 Months")),
                (365, _("Last 12 Months")),
                (0, _("All Time")),
            ],
            method=choice_noactivity,
        ),
        "with_account": wb_filters.BooleanFilter(
            label="With Account", help_text="Filter for entries without an account", method=_filter_with_account
        ),
        "with_account_role": wb_filters.BooleanFilter(
            label="With Account Role",
            help_text="Filter for entries without an account role",
            method=_filter_with_account_role,
        ),
        "interested_products": interested_products,
    }



from datetime import date, timedelta

import pytest
from django.contrib.auth.models import Permission
from faker import Faker
from psycopg.types.range import DateRange
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.directory.signals import deactivate_profile
from wbcore.utils.models import MergeError

from wbcrm.models import Account, AccountRole, AccountRoleValidity

fake = Faker()


@pytest.mark.django_db
class TestAccountModel:
    def test_init(self, account):
        assert account.id is not None

    def test_close_account_set_inactive(self, account):
        assert account.is_active
        account.status = Account.Status.CLOSE
        account.save()
        assert not account.is_active

    def test_disable_account_set_status_close(self, account):
        assert account.status == Account.Status.OPEN
        account.delete()
        assert not account.is_active
        assert account.status == Account.Status.CLOSE

    def test_account_with_owner_create_account_role(self, account_factory, entry_factory):
        customer = entry_factory.create()
        account = account_factory.create(owner=customer)
        assert account.roles.filter(entry=customer, role_type__key="customer").exists()

    def test_leaf_account_automatically_set_to_terminal(self, account_factory):
        parent_account = account_factory.create()

        assert parent_account.is_terminal_account is True
        child_account1 = account_factory.create(parent=parent_account, is_terminal_account=False)
        child_account2 = account_factory.create(parent=parent_account, is_terminal_account=False)

        parent_account.refresh_from_db()
        child_account1.refresh_from_db()
        child_account2.refresh_from_db()

        assert parent_account.is_terminal_account is False
        assert child_account1.is_terminal_account is True
        assert child_account2.is_terminal_account is True

        child_account1.delete(no_deletion=False)
        parent_account.refresh_from_db()
        assert parent_account.is_terminal_account is False  # because there is still another child

        child_account2.delete(no_deletion=False)
        parent_account.refresh_from_db()
        assert parent_account.is_terminal_account is True  # There is no children left so we expect it to become leaf

    def test_disabled_parent_account_disabled_children(self, account_factory):
        parent_account = account_factory.create()
        child_account = account_factory.create(parent=parent_account)
        assert child_account.is_active
        assert parent_account.is_active

        parent_account.is_active = False
        parent_account.save()

        child_account.refresh_from_db()
        assert not child_account.is_active

    def test_private_root_account_set_children_as_private(self, account_factory):
        parent_account = account_factory.create(is_public=True)
        child_account = account_factory.create(parent=parent_account, is_public=True)
        parent_account.is_public = False
        parent_account.save()
        child_account.refresh_from_db()
        assert not child_account.is_public

    def test_can_manage(self, account, user):
        assert not account.can_administrate(user)

        perm = Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
        user.user_permissions.add(perm)

        user = User.objects.get(id=user.id)  # we refetch user to clear perm cache, doesn't happen with refresh_from_db
        assert account.can_administrate(user)

    def test_get_inherited_roles_for_account(self, account_factory, account_role_factory):
        parent_account = account_factory.create()
        parent_role = account_role_factory.create(account=parent_account)
        child_account = account_factory.create(parent=parent_account)
        child_role = account_role_factory.create(account=child_account)
        assert set(child_account.get_inherited_roles_for_account()) == {parent_role}
        assert set(child_account.get_inherited_roles_for_account(include_self=True)) == {child_role, parent_role}

    @pytest.mark.parametrize("validity_date,account__is_public", [(fake.date_object(), False)])
    def test_can_see_account(self, user, account, account_role_factory, validity_date):
        # Check user doesn't see permission on the account by default
        assert not account.can_see_account(user, validity_date)
        role = account_role_factory.create(account=account, entry=Entry.objects.get(id=user.profile.id))
        # with a valid role, they can see it
        assert account.can_see_account(user, validity_date=validity_date)

        # with an unvalid role, they cannot see it
        role.validity_set.first().timespan = DateRange(date.min, validity_date - timedelta(days=1))
        role.validity_set.first().save()
        assert account.can_see_account(user, validity_date=validity_date)

    @pytest.mark.parametrize("account__is_public", [True, False])
    def test_can_see_account_with_internal_employee(self, internal_user_factory, account):
        # Check user doesn't see permission on the account by default

        # with being an internal user, they can see it
        assert account.can_see_account(internal_user_factory.create()) == account.is_public

    @pytest.mark.parametrize("account__is_public", [True])
    def test_normaluser_cannot_see_account_public_account(self, user, account):
        assert not account.can_see_account(user)

    def test_get_accounts_for_customer(self, account_factory, company_factory, employer_employee_relationship_factory):
        owner1 = company_factory.create()
        employee1 = employer_employee_relationship_factory.create(employer=owner1).employee
        owner2 = company_factory.create()
        employee2 = employer_employee_relationship_factory.create(employer=owner2).employee

        parent_account = account_factory.create(owner=owner1)
        other_parent_account = account_factory.create(owner=owner1)
        child_account = account_factory.create(parent=parent_account, owner=owner2)
        # test if all child account also shows in the queryset as well as the direct related account for that owner
        assert set(Account.get_accounts_for_customer(owner1.entry_ptr)) == {
            parent_account,
            other_parent_account,
            child_account,
        }
        assert set(Account.get_accounts_for_customer(employee1.entry_ptr)) == {
            parent_account,
            other_parent_account,
            child_account,
        }
        assert set(Account.get_accounts_for_customer(owner2.entry_ptr)) == {child_account}
        assert set(Account.get_accounts_for_customer(employee2.entry_ptr)) == {child_account}

    def test_get_managed_accounts_for_entry(self, entry, account_factory, account_role_factory):
        main_role = account_role_factory.create(entry=entry)
        child_account = account_factory.create(parent=main_role.account)
        other_account = account_factory.create()  # noqa
        other_role = account_role_factory.create()  # noqa
        assert set(Account.get_managed_accounts_for_entry(entry)) == {main_role.account, child_account}

    def test_handle_user_deactivation(self, account_role_factory, person_factory):
        profile = person_factory.create()
        substitute_profile = person_factory.create()

        valid_role = account_role_factory.create(entry=Entry.objects.get(id=profile.id))
        unvalid_role = account_role_factory.create(
            entry=Entry.objects.get(id=profile.id), visibility_daterange=DateRange(date.min, date.today())
        )

        deactivate_profile.send(AccountRole, instance=profile, substitute_profile=substitute_profile)
        valid_role.refresh_from_db()
        unvalid_role.refresh_from_db()
        assert valid_role.validity_set.first().timespan == DateRange(date.min, date.today() - timedelta(days=1))
        assert valid_role.entry.id == profile.id
        assert unvalid_role.entry.id == profile.id

        new_role = AccountRole.objects.get(entry_id=substitute_profile.id, account=valid_role.account)
        assert new_role.validity_set.first().timespan == DateRange(date.today() - timedelta(days=1), date.max)

    @pytest.mark.parametrize(
        "validity_date,is_internal_user, is_superuser",
        [
            (fake.date_object(), True, True),
            (fake.date_object(), False, True),
            (fake.date_object(), True, False),
            (fake.date_object(), False, False),
        ],
    )
    def test_filter_for_user(
        self,
        user_factory,
        internal_user_factory,
        account_factory,
        account_role_factory,
        validity_date,
        is_internal_user,
        is_superuser,
    ):
        # We assume role logic are tested in the AccountRole manager unit test
        # so we test here only:
        # - chain filtering
        # - Return account where user has a valid role
        # - and Return public account if user is superuser or internal

        if is_internal_user:
            user = internal_user_factory.create()
        else:
            user = user_factory.create()
        if is_superuser:
            user.user_permissions.add(
                Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
            )
        profile_entry = Entry.objects.get(id=user.profile.id)
        public_account = account_factory.create(is_public=True)
        private_account = account_factory.create(is_public=False)  # noqa
        valid_role = account_role_factory.create(account__is_public=fake.pybool(), entry=profile_entry)
        children_valid_role_account = account_factory.create(
            parent=valid_role.account, is_public=True
        )  # check that entry with role on the parent account can access the public children account
        account_factory.create(
            parent=valid_role.account, is_public=False
        )  # check that entry with role on the parent account cannot access private children account

        unvalid_role = account_role_factory.create(  # noqa
            account__is_public=False,
            entry=profile_entry,
            visibility_daterange=DateRange(date.min, validity_date - timedelta(days=1)),
        )
        user = User.objects.get(id=user.id)
        if is_superuser:
            assert set(Account.objects.filter_for_user(user, validity_date)) == set(Account.objects.all())
        elif is_internal_user:
            assert set(Account.objects.filter_for_user(user, validity_date)) == {
                public_account,
                valid_role.account,
                children_valid_role_account,
            }
            assert set(Account.objects.filter(id=public_account.id).filter_for_user(user, validity_date)) == {
                public_account
            }  # basic test filter chainage
        else:
            assert set(Account.objects.filter_for_user(user, validity_date)) == {
                valid_role.account,
                children_valid_role_account,
            }

    def test_cannot_merge_account_with_children(self, account_factory):
        """
        Check that we cannot merge an account that still has children
        """
        base_account = account_factory.create()
        merged_account = account_factory.create()
        children_merged_account = account_factory.create(parent=merged_account)
        with pytest.raises(MergeError):
            base_account.merge(merged_account)
        merged_account.refresh_from_db()
        assert merged_account

        # but can merge the children into the parent account
        merged_account.merge(children_merged_account)
        with pytest.raises(Account.DoesNotExist):
            children_merged_account.refresh_from_db()

    def test_cannot_merge_account_on_different_trees(self, account_factory):
        """
        Check that we cannot merge an account that still has children
        """
        base_account = account_factory.create(parent=account_factory.create())
        merged_account = account_factory.create(parent=account_factory.create())
        account_factory.create(parent=merged_account)
        with pytest.raises(MergeError):
            base_account.merge(merged_account)
        # assert the accounts weren't touched
        base_account.refresh_from_db()
        merged_account.refresh_from_db()
        assert base_account
        assert merged_account

    def test_account_merging(self, account_factory, account_role_factory, user):
        # TODO implemetns for commission
        pivot_date = date(2023, 1, 1)
        base_account = account_factory.create()
        base_account_role = account_role_factory.create(
            account=base_account, visibility_daterange=DateRange(date.min, pivot_date)
        )
        base_account_role_validity = base_account_role.validity_set.first()

        merged_account = account_factory.create()
        merged_account_role_overlapping = account_role_factory.create(
            account=merged_account, entry=base_account_role.entry, visibility_daterange=DateRange(pivot_date, date.max)
        )
        merged_account_role_overlapping.authorized_hidden_users.add(user)
        merged_account_role_validity = merged_account_role_overlapping.validity_set.first()

        merged_account_role_different = account_role_factory.create(account=merged_account)
        merged_account_role_different_validity = merged_account_role_different.validity_set.first()

        base_account.merge(merged_account)

        # test that the none overlapping role were forwarded to the base account
        assert set(base_account.roles.all()) == {base_account_role, merged_account_role_different}

        # test that the validity set for the base role got the new validity from the base merged account role
        assert set(base_account_role.validity_set.all()) == {base_account_role_validity, merged_account_role_validity}

        # test that the authorized user set on the merged account role were forwarded to the base account
        assert set(base_account_role.authorized_hidden_users.all()) == {user}

        # test that the merged account was deleted
        with pytest.raises(Account.DoesNotExist):
            merged_account.refresh_from_db()

        # test that the non overlapping role correctly follows the role
        role = AccountRole.objects.get(account=base_account, entry=merged_account_role_different.entry)
        assert set(role.validity_set.all()) == {merged_account_role_different_validity}


@pytest.mark.django_db
class TestAccountRoleModel:
    def test_init(self, account_role):
        assert account_role.id is not None
        validity = account_role.validity_set.first()
        assert validity
        assert validity.timespan == DateRange(date.min, date.max)

    def test_hidden_role_set_account_to_private(self, account_factory, account_role_factory):
        public_account = account_factory.create(is_public=True)
        role = account_role_factory.create(account=public_account, is_hidden=True)
        assert not role.account.is_public

    # def test_filter_for_user(self, ):
    #     # here we want to check that the chained filtering works and
    #     # - Internal or super user can see their roles + public account roles if there are unhidden + hidden role if they have the right to see them
    #     # - Otherwise, can only see direct roles
    #     # - Check if the annotation is_currently_valid is there
    @pytest.mark.parametrize(
        "is_superuser,is_internal_user",
        [
            (True, True),
            (True, False),
            (False, True),
            (False, False),
        ],
    )
    def test_filter_related_account_roles_for_internal_user(
        self, user_factory, internal_user_factory, account_role_factory, is_superuser, is_internal_user
    ):
        if is_internal_user:
            user = internal_user_factory.create()
        else:
            user = user_factory.create()
        if is_superuser:
            user.user_permissions.add(
                Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
            )
        role_unhidden_public_account = account_role_factory.create(is_hidden=False, account__is_public=True)
        role_unhidden_private_account = account_role_factory.create(is_hidden=False, account__is_public=False)  # noqa
        role_hidden_public_account_without_privilege = account_role_factory.create(  # noqa
            is_hidden=True, account__is_public=True
        )

        role_hidden_private_account_with_privilege = account_role_factory.create(
            is_hidden=True, account__is_public=False, authorized_hidden_users=[user]
        )
        role_direct = account_role_factory.create(
            entry=Entry.objects.get(id=user.profile.id),
            account__is_public=fake.pybool(),
        )
        role_indirect_private_account = account_role_factory.create(account__is_public=False)  # noqa

        if is_superuser:
            assert set(AccountRole.objects.filter_for_user(user)) == set(AccountRole.objects.all())
        elif is_internal_user:
            assert set(AccountRole.objects.filter_for_user(user)) == {
                role_unhidden_public_account,
                role_hidden_private_account_with_privilege,
                role_direct,
            }
            assert set(AccountRole.objects.filter(is_hidden=True).filter_for_user(user)) == {
                role_hidden_private_account_with_privilege
            }  # Test that chain filtering works
        else:
            assert set(AccountRole.objects.filter_for_user(user)) == {role_direct}


@pytest.mark.django_db
class TestAccountRoleValidityModel:
    @pytest.mark.parametrize("validity_date", [fake.date_object()])
    def test_get_role_validity_subquery(self, account_role, validity_date):
        assert AccountRole.objects.count() == 1
        assert (
            AccountRole.objects.annotate(is_valid=AccountRoleValidity.get_role_validity_subquery(validity_date))
            .filter(is_valid=True)
            .count()
            == 1
        )

        validity = account_role.validity_set.first()
        validity.timespan = DateRange(date.min, validity_date - timedelta(days=1))
        validity.save()
        assert (
            not AccountRole.objects.annotate(is_valid=AccountRoleValidity.get_role_validity_subquery(validity_date))
            .filter(is_valid=True)
            .exists()
        )



from celery import shared_task
from django.db.utils import OperationalError
from django.http import HttpRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from wbcore.workers import Queue

from wbcrm.models.events import Event

from .shortcuts import get_backend


@shared_task(
    queue=Queue.HIGH_PRIORITY.value,
    default_retry_delay=5,
    autoretry_for=(OperationalError,),
    max_retries=4,
    retry_backoff=True,
)
def handle_inbound_as_task(event: dict):
    """
    the events received from the webhook are handled in a task
    which will allow to create, modify or delete the activity without interrupting the main server
    """
    if controller := get_backend():
        event_object = Event.objects.create(data=event)
        controller.handle_inbound(event=event, event_object_id=event_object.id)


@csrf_exempt
def event_watch(request: HttpRequest) -> HttpResponse:
    # TODO this is unsecure as it is prone to DDOS attack
    status_code = 200
    try:
        if controller := get_backend():
            if response := controller.handle_inbound_validation_response(request):
                return response
            for event in controller.get_events_from_inbound_request(request):
                handle_inbound_as_task.delay(event)
    except Exception as e:
        print(e)  # noqa: T201
        status_code = 500
    return HttpResponse(status=status_code)



from typing import TYPE_CHECKING

from langchain_core.messages import HumanMessage, SystemMessage
from pydantic import BaseModel, Field
from wbcore.contrib.ai.llm.config import LLMConfig

if TYPE_CHECKING:
    from wbcrm.models import Activity  # noqa


class ActivityLLMResponseModel(BaseModel):
    heat: int = Field(..., ge=1, le=4, description="The sentiment heat.")
    summary: str = Field(..., description="A summary of the activity in English.")


def get_query(instance):
    return {
        "title": instance.title,
        "description": instance.description,
        "period": instance.period,
        "participants": instance.participants.all(),
        "companies": instance.companies.all(),
        "result": instance.result,
    }


analyze_activity = LLMConfig["Activity"](
    key="analyze",
    prompt=[
        SystemMessage(
            content="You are a manager in a company and there are activities, both internal as well as external, meaning between service providers, clients and ourselves. We want to analyze these activities in regards of their sentiment and summarize them in English. The summary should be a very short bottom-line focussed text."
        ),
        HumanMessage(
            content="title={title}, description={description}, period={period}, participants={participants}, companies={companies}, review={result}"
        ),
    ],
    on_save=True,
    on_condition=lambda instance: instance.status == "REVIEWED",
    output_model=ActivityLLMResponseModel,
    query=get_query,
)



from contextlib import suppress

from django.conf import settings
from django.utils.module_loading import import_string
from dynamic_preferences.exceptions import NotFoundInRegistry
from dynamic_preferences.registries import global_preferences_registry


def get_backend():
    from wbcrm.synchronization.activity.controller import ActivityController

    if not settings.DEBUG:
        with suppress(NotFoundInRegistry):
            if backend := global_preferences_registry.manager()["wbactivity_sync__sync_backend_calendar"]:
                backend = import_string(backend)
                return ActivityController(backend=backend)



from celery import shared_task
from wbcore.workers import Queue

from .shortcuts import get_backend


@shared_task(queue=Queue.DEFAULT.value)
def periodic_notify_admins_of_webhook_inconsistencies_task(emails: list | None = None):
    """
    Periodic tasks to notify webhook inconsistencies
    """
    if emails and (controller := get_backend()):
        controller.backend.notify_admins_of_webhook_inconsistencies(emails)


@shared_task(queue=Queue.DEFAULT.value)
def periodic_renew_web_hooks_task():
    """
    Periodic tasks to renew active webhooks
    """
    if controller := get_backend():
        controller.backend.renew_web_hooks()



import json
import re
import time
from datetime import date, timedelta
from typing import Any

from dateutil import parser
from dateutil.relativedelta import relativedelta
from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.authentication.models.users import User

from wbcrm.synchronization.activity.backend import SyncBackend
from wbcrm.synchronization.activity.utils import flattened_dict_into_nested_dict
from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import User as UserDTO

from .msgraph import MicrosoftGraphAPI
from .parser import OutlookParser, parse


class OutlookSyncBackend(SyncBackend):
    METADATA_KEY = "outlook"

    def open(self):
        self.msgraph = MicrosoftGraphAPI()

    def close(self):
        pass

    def _validation_response(self, request: HttpRequest) -> HttpResponse:
        # handle validation
        if token := request.GET.get("validationToken"):
            return HttpResponse(token, content_type="text/plain")
        # handle callback consent permision
        if _consent := request.GET.get("admin_consent"):
            return HttpResponse(_consent, content_type="text/plain")
        return super()._validation_response(request)

    def _is_inbound_request_valid(self, request: HttpRequest) -> bool:
        is_valid = False
        try:
            if request.body and (json_body := json.loads(request.body)) and (notifications := json_body.get("value")):
                # When many changes occur, MSGraph may send multiple notifications that correspond to different subscriptions in the same POST request.
                for notification in parse(notifications):
                    if (
                        notification.get("client_state")
                        == global_preferences_registry.manager()["wbactivity_sync__outlook_sync_client_state"]
                        and (resource_data := notification.get("resource_data"))
                        and (event_type := resource_data.get("@odata.type", "").lower())
                        and (event_type == "#microsoft.graph.event")
                    ):
                        is_valid = True
                    else:
                        return False
        except json.decoder.JSONDecodeError:
            return False
        return is_valid

    def _get_events_from_request(self, request: HttpRequest) -> list[dict[str, Any]]:
        events = []
        for notification in parse(json.loads(request.body)["value"]):
            event = {
                "change_type": notification["change_type"],
                "resource": notification["resource"],
                "subscription_id": notification["subscription_id"],
            }
            if data := self.msgraph.get_event_by_resource(notification["resource"]):
                # Try to get the organizer's event in case if it's an invitation event
                if data["is_organizer"] is True:
                    event["organizer_resource"] = notification["resource"]
                elif (
                    (tenant_id := self.msgraph.get_tenant_id(data["organizer.email_address.address"]))
                    and data["is_organizer"] is False
                    and (organizer_data := self.msgraph.get_event_by_uid(tenant_id, data["uid"]))
                ):
                    event["organizer_resource"] = f"Users/{tenant_id}/Events/{organizer_data['id']}"
                    data = organizer_data
                event.update(data)
            events.append(event)
        return events

    def _deserialize(self, event: dict[str, Any], include_metadata: bool = True) -> tuple[ActivityDTO, bool, UserDTO]:
        event_id = event.get("id")
        delete_notification = event.get("change_type") == "deleted"
        is_deleted = delete_notification or not event_id or event.get("is_cancelled") is True
        user_dto = UserDTO(metadata={self.METADATA_KEY: {"subscription": {"id": event.get("subscription_id")}}})
        if include_metadata:
            _metadata = {"event_uid": event["uid"], "event_id": event_id} if event_id else {}
            if resource := event.get("resource"):
                _metadata["resources"] = [resource]
            if organizer_resource := event.get("organizer_resource"):
                _metadata["organizer_resource"] = organizer_resource
            elif is_deleted and resource:
                _metadata["organizer_resource"] = resource
            metadata = {self.METADATA_KEY: _metadata}
        else:
            metadata = {}

        if event_id:
            start = OutlookParser.convert_string_to_datetime(event["start.date_time"], event["start.time_zone"])
            end = OutlookParser.convert_string_to_datetime(event["end.date_time"], event["end.time_zone"])
            if start == end:
                end += timedelta(seconds=1)
            period = TimestamptzRange(start, end)
            if event["is_all_day"] is True:
                period = OutlookParser.convert_to_all_day_period(period)
            participants, conference_room = OutlookParser.deserialize_participants(event)
            recurrence_dict = OutlookParser.deserialize_recurring_activities(event)
            if event.get("is_reminder_on"):
                reminder_choice = OutlookParser.convert_reminder_minutes_to_choice(
                    event.get("reminder_minutes_before_start")
                )
            else:
                reminder_choice = ActivityDTO.ReminderChoice.NEVER.name
            activity_dto = ActivityDTO(
                metadata=metadata,
                title=event["subject"] if event.get("subject") else "(No Subject)",
                period=period,
                description=event.get("body.content", ""),
                participants=participants,
                creator=OutlookParser.deserialize_person(
                    event["organizer.email_address.address"], event["organizer.email_address.name"]
                ),
                visibility=OutlookParser.convert_sensitivity_to_visibility(event.get("sensitivity")),
                reminder_choice=reminder_choice,
                is_cancelled=event.get("is_cancelled", False),
                all_day=event.get("is_all_day", False),
                online_meeting=event.get("is_online_meeting", False),
                location=event.get("location.display_name"),
                conference_room=conference_room[0] if conference_room else None,
                recurrence_end=recurrence_dict["recurrence_end"],
                recurrence_count=recurrence_dict["recurrence_count"],
                repeat_choice=recurrence_dict["repeat_choice"],
            )
            if (
                event["type"] == "seriesMaster"
                and (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email))
                and (
                    occurrences := self.msgraph.get_instances_event(
                        tenant_id, event["id"], start.date(), end.date() + relativedelta(years=10)
                    )
                )
            ):
                recurring_activities = []
                for occurrence in occurrences:
                    instance_dto, *_ = self._deserialize(occurrence, include_metadata=False)
                    instance_dto.recurrence_end = activity_dto.recurrence_end
                    instance_dto.recurrence_count = activity_dto.recurrence_count
                    instance_dto.repeat_choice = activity_dto.repeat_choice
                    instance_dto.metadata = {
                        self.METADATA_KEY: {
                            **metadata[self.METADATA_KEY],
                            "occurrence_id": occurrence["id"],
                            "occurrence_resource": f"Users/{tenant_id}/Events/{occurrence['id']}",
                        }
                    }
                    if (extension := self.msgraph.get_extension_event(tenant_id, occurrence["id"])) and (
                        act_id := extension.get("activity_id")
                    ):
                        instance_dto.id = act_id
                    recurring_activities.append(instance_dto)
                activity_dto.recurring_activities = recurring_activities
        else:
            activity_dto = ActivityDTO(metadata=metadata, title=event.get("subject", "(No Subject)"))
        activity_dto.delete_notification = delete_notification
        return activity_dto, is_deleted, user_dto

    def _serialize(self, activity_dto: ActivityDTO, created: bool = False) -> dict[str, Any]:
        attendees = OutlookParser.serialize_participants(activity_dto.participants)
        if activity_dto.conference_room:
            attendees.append(OutlookParser.serialize_conference_room(activity_dto.conference_room))

        activity_dict = {
            "subject": activity_dto.title,
            "start": {
                "dateTime": activity_dto.period.lower.isoformat(),
                "timeZone": activity_dto.period.lower.tzname(),
            },
            "end": {
                "dateTime": activity_dto.period.upper.isoformat(),
                "timeZone": activity_dto.period.upper.tzname(),
            },
            "body": {
                "contentType": "HTML",
                "content": activity_dto.description,
            },
            "attendees": attendees,
            "sensitivity": OutlookParser.convert_visibility_to_sensitivity(activity_dto.visibility),
            "reminderMinutesBeforeStart": OutlookParser.convert_reminder_choice_to_minutes(
                activity_dto.reminder_choice
            ),
            "isReminderOn": False if activity_dto.reminder_choice == ActivityDTO.ReminderChoice.NEVER.name else True,
            "isAllDay": activity_dto.all_day,
            "responseRequested": True,
        }

        if activity_dto.online_meeting:
            activity_dict.update(
                {"allowNewTimeProposals": True, "isOnlineMeeting": True, "onlineMeetingProvider": "teamsForBusiness"}
            )
        if activity_dto.location:
            activity_dict["location"] = {
                "displayName": (
                    activity_dto.location if isinstance(activity_dto.location, str) else str(activity_dto.location)
                )
            }
        activity_dict["locations"] = [{"displayName": activity_dto.location}] if activity_dto.location else []
        if created and activity_dto.is_recurrent:
            activity_dict.update(OutlookParser.serialize_recurring_activities(activity_dto))
        return activity_dict

    def _stream_deletion(self, activity_dto: ActivityDTO):
        if event := self.get_external_event(activity_dto):
            if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
                self.msgraph.delete_event(tenant_id, event["id"])

    def _stream_creation(
        self, activity_dto: ActivityDTO, activity_dict: dict[str, Any]
    ) -> list[tuple[ActivityDTO, dict[str, Any]]] | None:
        if not self.get_external_event(activity_dto) and (
            tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)
        ):
            if event := self.msgraph.create_event(tenant_id, activity_dict):
                metadata_list = self._get_metadata_from_event(activity_dto, event)
                for act_dto, _metadata in metadata_list:
                    if act_dto.is_recurrent:
                        self._stream_extension_event(act_dto, _metadata)
                return metadata_list

    def _stream_update(
        self,
        activity_dto: ActivityDTO,
        activity_dict: dict[str, Any] | None = None,
        only_participants_changed: bool = False,
        external_participants: list | None = None,
        keep_external_description: bool = False,
    ) -> tuple[ActivityDTO, dict[str, Any]]:
        if external_participants is None:
            external_participants = []
        if old_event := self.get_external_event(activity_dto):
            if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
                activity_dict = activity_dict or old_event
                # When a participant is removed, we send a complete update of the event to remove this person in outlook
                activity_dict["attendees"] += external_participants
                if only_participants_changed:
                    # An event update that includes only the attendees property in the request body sends a meeting update to only the attendees that have changed.
                    # An event update that removes an attendee specified as a member of a distribution list sends a meeting update to all the attendees.
                    activity_dict = {"attendees": activity_dict["attendees"]}
                elif keep_external_description:
                    # preserve body when description is not synchronized especially meeting blob for online meeting.
                    activity_dict["body"]["content"] = old_event.get("body.content", "")

                if (
                    activity_dto.is_recurrent
                    and activity_dto.propagate_for_all_children
                    and not activity_dto.is_leaf
                    and (old_master_event := self.get_external_event(activity_dto, master_event=True))
                ):
                    if activity_dto.is_root:
                        self.msgraph.update_event(tenant_id, old_master_event["id"], activity_dict)
                    else:
                        return self._stream_update_from_occurrence(activity_dto, old_master_event, activity_dict)
                else:
                    self.msgraph.update_event(tenant_id, old_event["id"], activity_dict)
        elif not activity_dto.metadata.get(self.METADATA_KEY):
            # creation of recurring activity is done only during the creation phase of the activity
            if activity_dto.is_recurrent and activity_dict.get("recurrence"):
                activity_dict.pop("recurrence")
            return self._stream_creation(activity_dto, activity_dict)

    def _stream_update_from_occurrence(self, activity_dto: ActivityDTO, old_master_event: dict, activity_dict: dict):
        if (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)) and (
            instances_dict := self._get_occurrences_from_master_event(
                old_master_event, start_date_as_key=True, from_date=activity_dto.period.lower.date()
            )
        ):
            # Update recurrence pattern from old root to new root
            remaining_old_recurrence = {
                "recurrence": flattened_dict_into_nested_dict(OutlookParser.serialize_event_keys(old_master_event))[
                    "recurrence"
                ]
            }
            last_start_old = activity_dto.period.lower - timedelta(days=1)
            if remaining_old_recurrence := OutlookParser.serialize_recurrence_range_with_end_date(
                remaining_old_recurrence, last_start_old
            ):
                self.msgraph.update_event(tenant_id, old_master_event["id"], remaining_old_recurrence)
            # Create recurrence from new root
            new_recurrence = OutlookParser.serialize_recurring_activities(activity_dto)
            last_start_new = sorted(instances_dict.keys()).pop()
            if new_recurrence := OutlookParser.serialize_recurrence_range_with_end_date(
                new_recurrence, last_start_new
            ):
                activity_dict.update(new_recurrence)
                if new_master_event := self.msgraph.create_event(tenant_id, activity_dict):
                    # delete occurrences that had been deleted in the past
                    if new_instances_dict := self._get_occurrences_from_master_event(
                        new_master_event, start_date_as_key=True
                    ):
                        for key, value in new_instances_dict.items():
                            if not instances_dict.get(key):
                                self.msgraph.delete_event(tenant_id, value["id"])
                    metadata_list = self._get_metadata_from_event(activity_dto, new_master_event)
                    for act_dto, _metadata in metadata_list:
                        if act_dto.is_recurrent:
                            self._stream_extension_event(act_dto, _metadata)
                    return metadata_list

    def _stream_update_only_attendees(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        new_attendees = OutlookParser.serialize_participants(participants_dto)
        self._stream_update(
            activity_dto=activity_dto,
            external_participants=new_attendees,
            only_participants_changed=True,
        )

    def _stream_forward(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        if (event := self.get_external_event(activity_dto)) and (
            tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)
        ):
            new_attendees = []
            participants, _ = OutlookParser.deserialize_participants(event)
            external_emails_participants = [participant.person.email for participant in participants]
            for participant_dto in participants_dto:
                # forward if it is not the organizer or if it is not already part of the participants
                if (
                    participant_dto.person.email != activity_dto.creator.email
                    and participant_dto.person.email not in external_emails_participants
                ):
                    new_attendees.append(OutlookParser.serialize_person(participant_dto.person))
                    self.msgraph.forward_event(tenant_id, event["id"], new_attendees)

    def _stream_participant_change(
        self, participant_dto: ParticipantStatusDTO, is_deleted: bool = False, wait_before_changing: bool = False
    ):
        if (organizer_event := self.get_external_event(participant_dto.activity)) and (
            tenant_id := self.msgraph.get_tenant_id(participant_dto.person.email)
        ):
            if wait_before_changing:
                time.sleep(10)
            if invitation_event := self.msgraph.get_event_by_uid(tenant_id, organizer_event["uid"]):
                if participant_dto.person.email == participant_dto.activity.creator.email:
                    if is_deleted or participant_dto.status == ParticipantStatusDTO.ParticipationStatus.CANCELLED.name:
                        pass  # self.msgraph.cancel_event(tenant_id, invitation_event["id"]) # delete activity if organizer is removed of participants
                else:
                    if is_deleted or participant_dto.status == ParticipantStatusDTO.ParticipationStatus.CANCELLED.name:
                        self.msgraph.decline_event(tenant_id, invitation_event["id"])
                    elif participant_dto.status == ParticipantStatusDTO.ParticipationStatus.MAYBE.name:
                        self.msgraph.tentatively_accept_event(tenant_id, invitation_event["id"])
                    elif participant_dto.status in [
                        ParticipantStatusDTO.ParticipationStatus.ATTENDS.name,
                        ParticipantStatusDTO.ParticipationStatus.ATTENDS_DIGITALLY.name,
                    ]:
                        self.msgraph.accept_event(tenant_id, invitation_event["id"])

    def _stream_extension_event(self, activity_dto: ActivityDTO, metadata: dict | None = None) -> None:
        metadata = metadata if metadata else activity_dto.metadata
        if (
            (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email))
            and (_metadata := metadata.get(self.METADATA_KEY))
            and (occurrence_id := _metadata.get("occurrence_id"))
        ):
            self.msgraph.update_or_create_extension_event(tenant_id, occurrence_id, {"activity_id": activity_dto.id})

    def _set_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        response = {}
        self.open()
        if (
            (outlook := user.metadata.get(self.METADATA_KEY))
            and (subscription := outlook.get("subscription"))
            and (ms_subscription := self.msgraph.subscription(subscription.get("id")))
        ):
            response = self.msgraph._renew_subscription(ms_subscription.get("id"))
        elif tenant_id := self.msgraph.get_tenant_id(user.email):
            response = self.msgraph._subscribe(f"users/{tenant_id}/events/", "created, updated, deleted")
        else:
            raise Exception(gettext("Outlook TenantId not found for: ") + str(user))
        self.close()
        return {"subscription": response} if response else user.metadata.get(self.METADATA_KEY, {})

    def _stop_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        if (
            (outlook_backend := user.metadata.get(self.METADATA_KEY))
            and (subscription := outlook_backend.get("subscription"))
            and (subscription_id := subscription.get("id"))
        ):
            self.open()
            self.msgraph._unsubscribe(subscription_id)
            self.close()
            del user.metadata[self.METADATA_KEY]["subscription"]
        else:
            raise ValueError(str(user) + gettext(" has no active webhook"))
        return user.metadata.get(self.METADATA_KEY)

    def check_web_hook(self, user: "User", raise_error: bool = True) -> bool:
        error = ""
        self.open()
        if (outlook := user.metadata.get(self.METADATA_KEY)) and (subscription := outlook.get("subscription")):
            if not self.msgraph.subscription(subscription.get("id")):
                error = gettext("Webhook is invalid, Remove or Stop it and Set again please. ")
        else:
            error = gettext("Webhook not found. ")
        if error and raise_error:
            if user_tenant_id := self.msgraph.get_tenant_id(user.email):
                tenant_ids = [
                    items[1]
                    for sub in self.msgraph.subscriptions()
                    if (resource := sub.get("resource")) and (items := resource.split("/")) and len(items) > 2
                ]
                error += gettext(
                    "Number of subscriptions found in outlook for {} out of the total number: {}/{}."
                ).format(user.email, tenant_ids.count(user_tenant_id), len(tenant_ids))
            else:
                error += gettext("TenantId not found for ") + str(user)
            raise Exception(error)
        self.close()
        return False if error else True

    def renew_web_hooks(self) -> None:
        self.open()
        lookup = {f"metadata__{self.METADATA_KEY}__subscription__id__isnull": False}
        for user in User.objects.filter(**lookup):
            if response := self.msgraph._renew_subscription(user.metadata[self.METADATA_KEY]["subscription"]["id"]):
                user.metadata[self.METADATA_KEY]["subscription"] = response
                user.save()
        self.close()

    def _get_webhook_inconsistencies(self) -> str:
        self.open()
        subscriptions = self.msgraph.subscriptions()
        calls = set(
            list(
                map(
                    lambda x: x["id"],
                    list(filter(lambda x: x["resource"] == "/communications/callRecords", subscriptions)),
                )
            )
        )
        calendars = set(
            list(
                map(
                    lambda x: x["id"],
                    list(filter(lambda x: bool(re.match(r"users\/.*\/events\/", x["resource"])), subscriptions)),
                )
            )
        )
        lookup = f"metadata__{self.METADATA_KEY}__subscription__id"
        stored_calendars = set(User.objects.filter(**{f"{lookup}__isnull": False}).values_list(lookup, flat=True))
        message = ""
        if len(calls) == 0:
            message += gettext_lazy("<li>No Call Record subscription found in Microsoft</li>")
        if calendars != stored_calendars:
            diff = calendars.difference(stored_calendars)
            diff_inv = stored_calendars.difference(calendars)
            message += gettext_lazy(
                """
                <li>Number of calendar subscription not found in our system : <b>{}</b></li>
                <p>{}</p>

                <li>Number of calendar subscription assumed to be active not found in outlook: <b>{}</b></li>
                <p>{}</p>
            """
            ).format(len(diff), diff, len(diff_inv), diff_inv)
        if message == "":
            for subscription in subscriptions:
                if expiration_date_time := subscription.get("expiration_date_time"):
                    date_time = parser.parse(expiration_date_time)
                    diff = date_time - timezone.now()
                    if diff.days < 1:
                        message += f"Resource: {subscription['resource']} expires on {date_time} (in {diff})"
        self.close()
        return message

    def get_external_event(self, activity_dto: ActivityDTO, master_event: bool = False) -> dict:
        event = None
        metadata = activity_dto.metadata.get(self.METADATA_KEY, {})
        if master_event or not activity_dto.is_recurrent:
            if resource := metadata.get("organizer_resource"):
                event = self.msgraph.get_event_by_resource(resource)
            if (creator := activity_dto.creator) and (tenant_id := self.msgraph.get_tenant_id(creator.email)):
                if not event and (event_id := metadata.get("event_id")):
                    event = self.msgraph.get_event(tenant_id, event_id)
                if not event and (event_uid := metadata.get("event_uid")):
                    event = self.msgraph.get_event_by_uid(tenant_id, event_uid)
        else:
            if resource := metadata.get("occurrence_resource"):
                event = self.msgraph.get_event_by_resource(resource)
            elif (
                not event
                and (creator := activity_dto.creator)
                and (tenant_id := self.msgraph.get_tenant_id(creator.email))
                and metadata.get("occurrence_id")
            ):
                event = self.msgraph.get_event(tenant_id, metadata["occurrence_id"])
        return event

    def get_external_participants(
        self, activity_dto: ActivityDTO, internal_participants_dto: list[ParticipantStatusDTO]
    ) -> list:
        external_participants = []
        if event := self.get_external_event(activity_dto):
            internal_emails = [
                participant_dto.person.email for participant_dto in internal_participants_dto if participant_dto.person
            ]
            for participant in event.get("attendees", []):
                if (
                    "address" in participant["emailAddress"]
                    and participant["emailAddress"]["address"] not in internal_emails
                ):
                    external_participants.append(participant)
        return external_participants

    def _is_participant_valid(self, user: "User") -> bool:
        try:
            return super()._is_participant_valid(user) and self.check_web_hook(user, raise_error=False)
        except Exception:
            return False

    def _generate_event_metadata(self, tenant_id: str, master_event: dict, occurrence_event: dict | None = None):
        if occurrence_event is None:
            occurrence_event = {}
        resource = f"Users/{tenant_id}/Events/{master_event['id']}"
        metadata = {
            self.METADATA_KEY: {
                "resources": [resource],
                "organizer_resource": resource,
                "event_uid": master_event["uid"],
                "event_id": master_event["id"],
            }
        }
        if occurrence_event:
            metadata[self.METADATA_KEY].update(
                {
                    "occurrence_id": occurrence_event["id"],
                    "occurrence_resource": f"Users/{tenant_id}/Events/{occurrence_event['id']}",
                }
            )
        return metadata

    def _get_metadata_from_event(
        self, activity_dto: ActivityDTO, event: dict
    ) -> list[tuple[ActivityDTO, dict[str, Any]]]:
        metadata_list = []
        if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
            if event.get("type") == "seriesMaster" and (
                instances_dict := self._get_occurrences_from_master_event(event, start_date_as_key=True)
            ):
                occurrence = instances_dict.get(activity_dto.period.lower.date(), {})
                metadata_list.append((activity_dto, self._generate_event_metadata(tenant_id, event, occurrence)))
                instances_dto = activity_dto.recurring_activities + activity_dto.invalid_recurring_activities
                for instance_dto in instances_dto:
                    if occurrence := instances_dict.get(instance_dto.period.lower.date()):
                        metadata_list.append(
                            (instance_dto, self._generate_event_metadata(tenant_id, event, occurrence))
                        )
            else:
                metadata_list.append((activity_dto, self._generate_event_metadata(tenant_id, event)))
        return metadata_list

    def _get_occurrences_from_master_event(
        self, event: dict, start_date_as_key: bool = False, from_date: date = None
    ) -> dict:
        """
        Dict of instances event whose key is id by default otherwise start date and the value is the occurrence event data.
        """
        start = (
            from_date
            if from_date
            else OutlookParser.convert_string_to_datetime(event["start.date_time"], event["start.time_zone"]).date()
        )
        end = OutlookParser.convert_string_to_datetime(
            event["end.date_time"], event["end.time_zone"]
        ).date() + relativedelta(years=10)
        tenant_id = self.msgraph.get_tenant_id(event["organizer.email_address.address"])
        occurrences_dict = {}
        if occurrences := self.msgraph.get_instances_event(tenant_id, event["id"], start, end):
            if start_date_as_key:
                occurrences_dict = {
                    OutlookParser.convert_string_to_datetime(
                        occurrence["start.date_time"], occurrence["start.time_zone"]
                    ).date(): occurrence
                    for occurrence in occurrences
                }
            else:
                occurrences_dict = {occurrence["id"]: occurrence for occurrence in occurrences}
        return occurrences_dict



from .accounts import (
    AccountModelViewSet,
    AccountRepresentationViewSet,
    AccountRoleAccountModelViewSet,
    AccountRoleTypeRepresentationViewSet,
    ChildAccountAccountModelViewSet,
    InheritedAccountRoleAccountModelViewSet,
    AccountRoleValidityModelViewSet
)
from .activities import (
    ActivityChartModelViewSet,
    ActivityParticipantModelViewSet,
    ActivityRepresentationViewSet,
    ActivityTypeModelViewSet,
    ActivityTypeRepresentationViewSet,
    ActivityViewSet,
)
from .groups import GroupModelViewSet, GroupRepresentationViewSet
from .products import (
    ProductCompanyRelationshipCompanyModelViewSet,
    ProductModelViewSet,
    ProductRepresentationViewSet,
)



# Generated by Django 5.2.9 on 2025-12-09 08:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0014_alter_entry_relationship_managers_and_more'),
        ('wbcrm', '0018_activity_search_vector'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='companies',
            field=models.ManyToManyField(blank=True, help_text='The list of companies other than the main company that participate in this activity', related_name='company_participates', through='wbcrm.ActivityCompanyThroughModel', through_fields=('activity', 'company'), to='directory.company', verbose_name='Participating Companies'),
        ),
        migrations.AlterField(
            model_name='activity',
            name='participants',
            field=models.ManyToManyField(blank=True, help_text='The list of participants', related_name='participates', through='wbcrm.ActivityParticipant', through_fields=('activity', 'participant'), to='directory.person', verbose_name='Participating Persons'),
        ),
        migrations.AlterField(
            model_name='product',
            name='prospects',
            field=models.ManyToManyField(blank=True, help_text='The list of prospects', related_name='interested_products', through='wbcrm.ProductCompanyRelationship', through_fields=('product', 'company'), to='directory.company', verbose_name='Prospects'),
        ),
    ]



import factory
from slugify import slugify

from wbcrm.models.accounts import Account, AccountRole, AccountRoleType


class AccountFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("company")
    status = Account.Status.OPEN
    # owner =
    is_public = True
    is_active = True
    is_terminal_account = True

    class Meta:
        model = Account


class AccountWithOwnerFactory(AccountFactory):
    owner = factory.SubFactory("wbcore.contrib.directory.factories.entries.EntryFactory")


class AccountRoleTypeFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("word")
    key = factory.LazyAttribute(lambda o: slugify(o.title))

    class Meta:
        model = AccountRoleType
        django_get_or_create = ["key"]


class AccountRoleFactory(factory.django.DjangoModelFactory):
    role_type = factory.SubFactory("wbcrm.factories.AccountRoleTypeFactory")
    entry = factory.SubFactory("wbcore.contrib.directory.factories.entries.EntryFactory")
    account = factory.SubFactory("wbcrm.factories.AccountFactory", parent=None)
    is_hidden = False

    @factory.post_generation
    def authorized_hidden_users(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for user in extracted:
                self.authorized_hidden_users.add(user)

    @factory.post_generation
    def visibility_daterange(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            v = self.validity_set.first()
            v.timespan = extracted
            v.save()

    class Meta:
        model = AccountRole
        skip_postgeneration_save = True



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.contrib.messages import info, warning
from django.db.models import Exists, F, OuterRef, QuerySet, Subquery
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.agenda.viewsets import CalendarItemViewSet
from wbcore.contrib.directory.models import (
    Company,
    EmailContact,
    EmployerEmployeeRelationship,
    Entry,
    Person,
    TelephoneContact,
)
from wbcore.filters import DjangoFilterBackend
from wbcore.utils.urls import get_multi_request_value

from wbcrm import serializers as crm_serializers
from wbcrm.filters import (
    ActivityChartFilter,
    ActivityFilter,
    ActivityParticipantFilter,
    ActivityTypeFilter,
)
from wbcrm.models.activities import (
    Activity,
    ActivityParticipant,
    ActivityType,
    send_invitation_participant_as_task,
)
from wbcrm.viewsets.buttons import ActivityButtonConfig, ActivityParticipantButtonConfig
from wbcrm.viewsets.display import (
    ActivityDisplay,
    ActivityParticipantDisplayConfig,
    ActivityTypeDisplay,
)
from wbcrm.viewsets.endpoints import (
    ActivityEndpointConfig,
    ActivityParticipantModelEndpointConfig,
)
from wbcrm.viewsets.previews.activities import ActivityPreviewConfig
from wbcrm.viewsets.titles import (
    ActivityChartModelTitleConfig,
    ActivityParticipantTitleConfig,
    ActivityTitleConfig,
    ActivityTypeTitleConfig,
)

from ..serializers import (
    ActivityTypeModelSerializer,
    ActivityTypeRepresentationSerializer,
)
from .recurrence import RecurrenceModelViewSetMixin


class ActivityTypeModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbcrm:activitytype"
    LIST_DOCUMENTATION = "wbcrm/markdown/documentation/activitytype.md"
    queryset = ActivityType.objects.all()
    serializer_class = ActivityTypeModelSerializer
    display_config_class = ActivityTypeDisplay
    title_config_class = ActivityTypeTitleConfig
    search_fields = ("title",)
    filterset_class = ActivityTypeFilter
    ordering = ("title",)
    ordering_fields = (
        "title",
        "color",
        "score",
        "icon",
        "default",
    )


class ActivityTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcrm:activitytyperepresentation"
    serializer_class = ActivityTypeRepresentationSerializer
    queryset = ActivityType.objects.all()


class ActivityRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Activity.objects.all()
    serializer_class = crm_serializers.ActivityRepresentationSerializer


class ActivityViewSet(RecurrenceModelViewSetMixin, CalendarItemViewSet):
    IDENTIFIER = "wbcrm:activity"
    DEPENDANT_IDENTIFIER = ["wbcrm:calendaritem"]
    LIST_DOCUMENTATION = "wbcrm/markdown/documentation/activity.md"

    ordering = ["-edited", "id"]
    search_fields = ("search_vector",)
    serializer_class = crm_serializers.ActivityModelSerializer
    display_config_class = ActivityDisplay
    title_config_class = ActivityTitleConfig
    filterset_class = ActivityFilter
    endpoint_config_class = ActivityEndpointConfig
    preview_config_class = ActivityPreviewConfig
    button_config_class = ActivityButtonConfig
    queryset = Activity.objects.all()

    ordering_fields = [
        "created",
        "description",
        "edited",
        "period",
        "latest_reviewer__computed_str",
        "type",
        "result",
        "title",
    ]

    @cached_property
    def is_external_activity(self):
        if "pk" in self.kwargs:
            if activity := self.get_object():
                return (creator := activity.creator) and not creator.is_internal
        return False

    @cached_property
    def is_private_for_user(self):
        if "pk" in self.kwargs:
            if activity := self.get_object():
                return activity.is_private_for_user(self.request.user)
        return not self.new_mode

    @cached_property
    def is_confidential_for_user(self):
        if "pk" in self.kwargs:
            if activity := self.get_object():
                return activity.is_confidential_for_user(self.request.user)
        return not self.new_mode

    @cached_property
    def participants(self) -> QuerySet[Person]:
        participant_ids = get_multi_request_value(self.request, "participants")
        return Person.objects.filter(id__in=participant_ids)

    @cached_property
    def companies(self) -> QuerySet[Company]:
        try:
            company_ids = get_multi_request_value(self.request, "companies")
        except KeyError:
            company_ids = []
        return Company.objects.filter(id__in=company_ids)

    @cached_property
    def entry(self) -> Entry | None:
        try:
            return Entry.all_objects.get(id=self.kwargs.get("person_id", self.kwargs["company_id"]))
        except (KeyError, Entry.DoesNotExist):
            return self.participants.first() or self.companies.first()

    def get_serializer_class(self):
        if self.get_action() in ["list", "list-metadata"]:
            return crm_serializers.ActivityModelListSerializer
        if (
            self.is_private_for_user
            or self.is_confidential_for_user
            or (
                "pk" in self.kwargs
                and (obj := self.get_object())
                and obj.status in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]
            )
        ):
            return crm_serializers.ReadOnlyActivityModelSerializer
        return super().get_serializer_class()

    def get_queryset(self):
        user = self.request.user
        return (
            Activity.get_activities_for_user(user, base_qs=super().get_queryset())
            .select_related("latest_reviewer", "type")
            .prefetch_related("groups", "participants", "companies", "activity_companies")
        )

    def add_messages(self, request, instance: Activity | None = None, **kwargs):
        super().add_messages(request, instance=instance, **kwargs)
        if instance:
            if instance.status in [Activity.Status.CANCELLED, Activity.Status.REVIEWED]:
                info(
                    self.request,
                    _(
                        "You can only modify the review text for an activity that is either cancelled or already reviewed."
                    ),
                )
            if self.is_external_activity:
                warning(
                    self.request,
                    _(
                        "This activity was created by an external user and synchronized. The modification is restricted."
                    ),
                )
            if warning_message := instance.participants_company_check_message():
                warning(request, warning_message, extra_tags="auto_close=0")

            # Throws a warning message when there are more people (probably) participating in an activity on-site
            # than the selected conference room has capacity for.
            if (
                instance.conference_room
                and instance.conference_room.capacity is not None
                and instance.participants.exclude(
                    activity_participants__participation_status__in=[
                        ActivityParticipant.ParticipationStatus.ATTENDS_DIGITALLY,
                        ActivityParticipant.ParticipationStatus.CANCELLED,
                    ]
                ).count()
                > instance.conference_room.capacity
            ):
                warning(
                    request,
                    _(
                        "There are more participants currently participating in this activity than the maximum capacity of the selected conference room allows."
                    ),
                )


class ActivityChartModelViewSet(viewsets.ChartViewSet):
    filter_backends = (DjangoFilterBackend,)
    queryset = Activity.objects.all()
    filterset_class = ActivityChartFilter
    IDENTIFIER = "wbcrm:activitychart"
    title_config_class = ActivityChartModelTitleConfig

    def get_queryset(self):
        return (
            Activity.get_activities_for_user(self.request.user)
            .exclude(period__isnull=True)
            .filter(visibility=CalendarItem.Visibility.PUBLIC)
            .annotate(
                activity_type_color=F("type__color"),
                activity_type_title=F("type__title"),
                start_date=F("period__startswith"),
                end_date=F("period__endswith"),
            )
            .select_related("type")
        )

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.exists():
            df = pd.DataFrame(
                queryset.values("type", "start_date", "end_date", "activity_type_color", "activity_type_title")
            )
            df["start_date"] = df["start_date"].dt.floor("h")
            df["end_date"] = df["end_date"].dt.ceil("h")
            df = (
                pd.concat(
                    [
                        pd.DataFrame(index=pd.date_range(r.start_date, r.end_date, freq="1h")).assign(
                            type=r.type,
                            activity_type_color=r.activity_type_color,
                            activity_type_title=r.activity_type_title,
                        )
                        for r in df.itertuples()
                    ]
                ).reset_index()
            ).rename(columns={"index": "Period", "activity_type_title": "Type"})
            df["Period"] = pd.to_datetime(df["Period"])
            fig = px.histogram(
                df,
                x="Period",
                color="Type",
                labels="Type",
                nbins=len(pd.date_range(df["Period"].min(), df["Period"].max(), freq="1h")) + 1,
            )
            fig.update_layout(
                paper_bgcolor="rgba(0,0,0,0)",
                plot_bgcolor="rgba(0,0,0,0)",
                title=_("<b>User Activity</b>"),
                bargap=0.2,
                autosize=True,
            )

        return fig


class ActivityParticipantModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbcrm:activity-participant"

    button_config_class = ActivityParticipantButtonConfig
    display_config_class = ActivityParticipantDisplayConfig
    endpoint_config_class = ActivityParticipantModelEndpointConfig
    filterset_class = ActivityParticipantFilter
    serializer_class = crm_serializers.ActivityParticipantModelSerializer
    search_fields = []
    title_config_class = ActivityParticipantTitleConfig
    ordering = ["participation_status", "id"]
    queryset = ActivityParticipant.objects.all()

    def get_queryset(self):
        activity = get_object_or_404(Activity, id=self.kwargs["activity_id"])
        return (
            super()
            .get_queryset()
            .filter(activity=activity)
            .annotate(
                customer_status=Subquery(
                    EmployerEmployeeRelationship.objects.filter(
                        primary=True, employee=OuterRef("participant__pk")
                    ).values("employer__customer_status__title")[:1]
                ),
                position=Subquery(
                    EmployerEmployeeRelationship.objects.filter(
                        primary=True, employee=OuterRef("participant__pk")
                    ).values("position__title")[:1]
                ),
                primary_telephone=Subquery(
                    TelephoneContact.objects.filter(primary=True, entry__id=OuterRef("participant__pk")).values(
                        "number"
                    )[:1],
                ),
                primary_email=Subquery(
                    EmailContact.objects.filter(primary=True, entry__id=OuterRef("participant__pk")).values("address")[
                        :1
                    ],
                ),
                is_occupied=Exists(
                    Activity.objects.filter(
                        period__overlap=activity.period,
                        participants__id=OuterRef("participant__pk"),
                    )
                    .exclude(id=activity.id)
                    .exclude(status=Activity.Status.CANCELLED)
                ),
            )
        )

    @action(methods=["PATCH"], detail=False)
    def send_external_invitation(self, request, activity_id: int, pk=None):
        send_invitation_participant_as_task.delay(activity_id)
        return Response({"__notification": {"title": "Invitation sent to external participants"}})



from wbcore.menus import ItemPermission, MenuItem

ACCOUNT_MENUITEM = MenuItem(
    label="Accounts",
    endpoint="wbcrm:account-list",
    endpoint_get_parameters={"parent__isnull": True},
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbcrm.view_account"]),
    add=MenuItem(
        label="Create Account",
        endpoint="wbcrm:account-list",
        permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbcrm.add_account"]),
    ),
)



from wbcore.tests.conftest import *  # isort:skip
import os

from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.agenda.factories import CalendarItemFactory, ConferenceRoomFactory
from wbcore.contrib.authentication.factories import (
    InternalUserFactory,
    SuperUserFactory,
    UserActivityFactory,
    UserFactory,
)
from wbcore.contrib.directory.factories import (
    CompanyFactory,
    EmailContactFactory,
    EmployerEmployeeRelationshipFactory,
    EntryFactory,
    PersonFactory,
)
from wbcrm.factories import (
    AccountFactory,
    AccountRoleFactory,
    AccountRoleTypeFactory,
    ActivityCompanyFactory,
    ActivityFactory,
    ActivityParticipantFactory,
    ActivityPersonFactory,
    ActivityTypeFactory,
    GroupFactory,
    ProductFactory,
    RecurringActivityFactory,
)

register(EntryFactory)
register(InternalUserFactory, "internal_user")
register(EmailContactFactory)
register(CompanyFactory)
register(PersonFactory)
register(EmployerEmployeeRelationshipFactory)
register(ActivityFactory)
register(ActivityTypeFactory)
register(RecurringActivityFactory)
register(ActivityCompanyFactory)
register(ActivityPersonFactory)
register(ActivityParticipantFactory)
register(GroupFactory)
register(ProductFactory)
register(ConferenceRoomFactory)
register(AccountFactory)
register(AccountRoleFactory)
register(AccountRoleTypeFactory)
register(CalendarItemFactory)

# Authentication
register(UserFactory)
register(SuperUserFactory, "superuser")
register(UserActivityFactory)


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True


@pytest.fixture
def chrome_options(chrome_options):
    # chrome_options.add_argument("--headless=new")
    return chrome_options


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



import datetime
import random

import factory
import pytz
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from wbcore.contrib.authentication.factories import InternalUserFactory
from wbcore.contrib.directory.factories import CompanyFactory, PersonFactory

from wbcrm.models.activities import Activity, ActivityParticipant, ActivityType

fake = Faker()


class ActivityTypeFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)
    color = factory.Faker("color")
    score = factory.Iterator([ActivityType.Score.HIGH, ActivityType.Score.MEDIUM, ActivityType.Score.LOW])

    class Meta:
        model = ActivityType
        django_get_or_create = ("title",)


class ActivityTypeCALLFactory(ActivityTypeFactory):
    title = "Call"


class ActivityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Activity
        skip_postgeneration_save = True

    title = factory.Faker("text", max_nb_chars=64)
    description = factory.Faker("paragraph", nb_sentences=5)
    start = factory.LazyAttribute(lambda _: fake.date_time(tzinfo=pytz.utc))
    end = factory.LazyAttribute(
        lambda _self: _self.start + datetime.timedelta(days=fake.pyint(0, 100), hours=fake.pyint(1, 23))
    )
    disable_participant_check = True
    location = factory.Faker("local_latlng")
    location_longitude = factory.Faker("longitude")
    location_latitude = factory.Faker("latitude")
    created = factory.Faker("date_time", tzinfo=pytz.utc)
    creator = factory.LazyAttribute(lambda _: InternalUserFactory.create().profile)
    latest_reviewer = factory.SubFactory(PersonFactory)
    reviewed_at = factory.Faker("date_time", tzinfo=pytz.utc)
    edited = factory.Faker("date_time", tzinfo=pytz.utc)
    assigned_to = factory.SubFactory(PersonFactory)
    preceded_by = None
    parent_occurrence = None
    propagate_for_all_children = False
    recurrence_end = None
    recurrence_count = None
    repeat_choice = Activity.ReoccuranceChoice.NEVER
    type = factory.SubFactory(ActivityTypeFactory)
    item_type = "wbcrm.Activity"

    @factory.post_generation
    def participants(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for participant in extracted:
                self.participants.add(participant)

    @factory.post_generation
    def companies(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for company in extracted:
                self.companies.add(company)

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for group in extracted:
                self.groups.add(group)

    @factory.post_generation
    def entities(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for entity in extracted:
                self.entities.add(entity)


class RecurringActivityFactory(ActivityFactory):
    repeat_choice = random.choice(list(filter(lambda x: x != "NEVER", Activity.ReoccuranceChoice.names)))
    recurrence_count = 3


class ActivityCompanyFactory(ActivityFactory):
    @factory.post_generation
    def companies(self, create, extracted, **kwargs):
        # Create company
        company = CompanyFactory()
        # Set global config main_company=company.id
        global_preferences_registry.manager()["directory__main_company"] = company.id
        self.companies.add(company)


class ActivityPersonFactory(ActivityFactory):
    @factory.post_generation
    def participants(self, create, extracted, **kwargs):
        # Create person
        person = PersonFactory()
        self.participants.add(person)


class ActivityParticipantFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ActivityParticipant

    participant = factory.SubFactory(PersonFactory)
    activity = factory.SubFactory(ActivityFactory)



from django.db import models
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from slugify import slugify
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.utils.models import ComplexToStringMixin


class ProductCompanyRelationship(models.Model):
    product = models.ForeignKey(
        on_delete=models.CASCADE,
        to="wbcrm.Product",
        verbose_name=_("Product"),
        related_name="product_company_relationships",
    )
    company = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Company",
        verbose_name=_("Company"),
        related_name="product_company_relationships",
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(name="unique_company_product_relationship", fields=["product", "company"])
        ]
        verbose_name = _("Company-Product Relationship")
        verbose_name_plural = _("Company-Product Relationships")

    def __str__(self) -> str:
        return f"{self.product} - {self.company}"


class Product(ComplexToStringMixin, WBModel):
    title = models.CharField(
        max_length=128,
        verbose_name=_("Title"),
    )

    slugify_title = models.CharField(
        max_length=128,
        verbose_name="Slugified Title",
        blank=True,
        null=True,
    )
    is_competitor = models.BooleanField(
        verbose_name=_("Is Competitor"),
        default=False,
        help_text=_("Indicates wether this is a competitor's product"),
    )

    prospects = models.ManyToManyField(
        "directory.Company",
        related_name="interested_products",
        blank=True,
        verbose_name=_("Prospects"),
        help_text=_("The list of prospects"),
        through="wbcrm.ProductCompanyRelationship",
        through_fields=("product", "company"),
    )

    def __str__(self) -> str:
        return self.title

    def compute_str(self) -> str:
        return _("{} (Competitor)").format(self.title) if self.is_competitor else self.title

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcrm:product"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcrm:productrepresentation-list"

    class Meta:
        verbose_name = _("Product")
        verbose_name_plural = _("Products")
        constraints = (models.UniqueConstraint(name="unique_product", fields=("slugify_title", "is_competitor")),)

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


@receiver(pre_merge, sender="directory.Company")
def pre_merge_entry_product(sender: models.Model, merged_object, main_object, **kwargs):
    for rel in ProductCompanyRelationship.objects.filter(company=merged_object):
        if ProductCompanyRelationship.objects.filter(company=main_object, product=rel.product).exists():
            rel.delete()
        else:
            rel.company = main_object
            rel.save()
